from dataclasses import field
from typing import Sequence, Union

import numpy as np
import pandas as pd

from oemof.solph._plumbing import sequence
from oemof.solph.buses import Bus
from oemof.solph.components import Converter
from oemof.solph.flows import Flow

import dataclasses
from oemof.tabular._facade import dataclass_facade, Facade
from oemof_tabular_plugins.wefe.facades import MIMO
from oemof_tabular_plugins.wefe.global_specs import crop_dict, soil_dict


@dataclass_facade
# @dataclasses.dataclass(unsafe_hash=False, frozen=False, eq=False)
class SimpleCrop(Converter, Facade):
    r"""Crop Model Converter with one input and one output. The crop growth factor
    is calculated out drought, heat, temperature, and water availabilty impact
     and considered for biomass production calculation.

    Parameters
    ----------
    crop_type: str
        the name of crop as defined in global_specs/crop_specs.py
    solar_bus: oemof.solph.Bus
        An oemof bus instance where the SimpleCrop unit is connected to with
        its input, it is expected to provide W/m² irradiance.
    harvest_bus: oemof.solph.Bus
        An oemof bus instance where the PV panel unit is connected to with
        its crop-harvest output. The unit is kg
    biomass_bus: oemof.solph.Bus
        An oemof bus instance where the SimpleCrop is connected with its
        non-edible biomass output. The unit is kg
    capacity: numeric
        The capacity of crop. It is expressed in cultivated area [m²]
    marginal_cost: numeric
        Marginal cost for one unit of produced output. Default: 0
    carrier_cost: numeric
        Carrier cost for one unit of used input. Default: 0
    capacity_cost: numeric
        Investment costs per unit of output capacity.
        If capacity is not set, this value will be used for optimizing the
        conversion output capacity.
    expandable: boolean or numeric (binary)
        True, if capacity can be expanded within optimization. Default: False.
    capacity_potential: numeric
        Maximum invest capacity in unit of output capacity.
    capacity_minimum: numeric
        Minimum invest capacity in unit of output capacity.
    input_parameters: dict (optional)
        Set parameters on the input edge of the conversion unit
        (see oemof.solph for more information on possible parameters)
    output_parameters: dict (optional)
        Set parameters on the output edge of the conversion unit
         (see oemof.solph for more information on possible parameters)

    ghi: time series
        Global horizontal irradiance
    et_0: time series
        potential evapotranspiration [m³]

    t_air: time series
        Ambient air temperature

    vwc: time series
        the voluemtric water content (vwc) in the root zone depth in m³; metric to express soil moisture


    SHOULD INCLUDE FUNCTIONS AND EXAMPLE HERE

    # TODO these parameters below are most likely provided by crop_dict

    light_saturation_point: numeric
        [lux]
    HI: numeric
        Harvest Index; share of biomass which is harvested as actual fruit
    I50A: numeric
        Irradiation fitting parameter
    I50B: numeric
        Irradation fitting parameter
    t_opt: numeric
         optimal temperature for biomass growth
    t_base: numeric
        base temperature for phenological development and growth
    rue: numeric
        Radiation Use efficiency (above ground only and without respiration) (g/MJm²)
    t_heat: numeric
        t_heat is the threshold temperature when biomass growth rate starts to be reduced by heat stress
    t_extreme: numeric
        t_ext is the extreme temperature threshold when the biomass growth rate reaches 0 due to heat stress
    t_sum: numeric
        cumulative temperature until harvest
    s_water: numeric
        sensitivity of RUE to the ARID index for specific plant (simple crop model)
    rzd: numeric
        root zone depth [m]

    """

    solar_bus: Bus

    harvest_bus: Bus

    carrier: str

    tech: str

    biomass_bus: Bus = None

    capacity: float = None

    marginal_cost: float = 0

    carrier_cost: float = 0

    capacity_cost: float = None

    expandable: bool = False

    capacity_potential: float = float("+inf")

    capacity_minimum: float = None

    input_parameters: dict = field(default_factory=dict)

    output_parameters: dict = field(default_factory=dict)

    crop_type: str = ""

    time_index: Union[float, Sequence[float]] = None

    t_air: Union[float, Sequence[float]] = None

    ghi: Union[float, Sequence[float]] = None

    et_0: Union[float, Sequence[float]] = None

    vwc: Union[float, Sequence[float]] = None

    sowing_date: str = ""  # TODO: MM-DD format

    harvest_date: str = ""  # TODO: MM-DD format

    @classmethod
    def validate_datapackage(self, resource):
        # check that the crop_type matches the one provided in the crop_dict
        data = pd.DataFrame.from_records(resource.read(keyed=True))

        for i, row in data[["crop_type", "name", "type"]].iterrows():
            if row.crop_type not in crop_dict:

                raise KeyError(
                    f"On line {i+1} of resource '{resource.descriptor['path']}', the crop_type attribute '{row.crop_type}' of the component of type '{row.type}', named '{row['name']}', is not available in the crop specs dict (wefe/gloabal_specs/crop_specs.py). Available crop types are: {', '.join([k for k in crop_dict.keys()])}"
                )
        return resource

    @classmethod
    def processing_raw_inputs(self, resource, results_df):
        # function to apply on df from above

        return results_df

    # @classmethod
    # def validate_datapackage(self, resource):
    #     # modify the resource (datapackage.resource)
    #     # should it return the resource?
    #     pass
    def get_crop_param(self, param_name):
        return crop_dict[self.crop_type][param_name]

    # efficiency equals biomass production factor calculate in the facade crop_specs.py; caapcity equals area [m²]

    def calc_Fsolar(
        self,
        time_index,
        sowing_date,
        harvest_date,
        t_air,
        t_base,
        t_sum,
        i50a,
        i50b,
        f_solar_max,
        **kwargs,
    ):
        """Calculates f_solar, the factor for interception of incoming solar radiation by plant canopy
            according to development stage [-].
            ----
        Parameters
        ----------
        time_index: time as pd.series or list
        sowing_date: time of sowing/start of cultivation period as str in "MM-DD" format
        harvest_date: time of harvest/end of cultivation period as str in "MM-DD" format
        t_air: ambient temperature as pd.series or list
        t_base: base temperature for biomass growth
        t_sum: cumulative temperature required from sowing to maturity
        i50a: cumulative temperature required from sowing on to reach 50% of solar radiation interception during growth
        i50b: cumulative temperature required until maturity to fall back to 50% of solar interception during decline
        f_solar_max: maximum factor for solar interception

        Returns
        -------
        f_solar_list : list of numerical values:
             solar interception coefficients for calculating biomass rate

        Notes
        -------
        This resembles the plant growth curve displayed in Fig 1.(e) of https://doi.org/10.1016/j.eja.2019.01.009

        """
        # Check input time series compatibility
        if not isinstance(t_air, (list, pd.Series)):
            print("Argument 'temp' is not of type list or pd.Series!")
        if not isinstance(time_index, (list, pd.Series)):
            print("Argument 'time index' is not of type list or pd.Series!")
        if len(t_air) != len(time_index):
            raise ValueError("Length mismatch between t_air and time_index profiles.")

        # Convert dates to Timestamp objects matching the time index
        dates = list(pd.to_datetime(time_index))
        year = str(dates[0].year)
        timezone = dates[0].tz
        # Opt. 1: sowing_date and harvest date given, plant maturity (t_sum) will be updated (custom_harvest=True)
        if sowing_date and harvest_date:
            sowing_date = pd.Timestamp(year + "-" + sowing_date).tz_localize(timezone)
            harvest_date = pd.Timestamp(year + "-" + harvest_date).tz_localize(timezone)
            custom_harvest = True
            # If harvest and sowing date are the same, move harvest date one time step back to avoid problems
            if sowing_date == harvest_date:
                harvest_date = dates[dates.index(harvest_date) - 1]
        # Opt. 2: only sowing date, harvest_date (end of cultivation period) is one day before (following year),
        # maturity according to SIMPLE
        elif sowing_date and not harvest_date:
            sowing_date = pd.Timestamp(year + "-" + sowing_date).tz_localize(timezone)
            harvest_date = dates[dates.index(sowing_date) - 1]
            custom_harvest = False
        # Opt. 3: no dates, growth from start of the year till maturity (from SIMPLE)
        else:
            sowing_date = dates[0]
            harvest_date = dates[-1]
            custom_harvest = False

        print(
            f"sowing date: {sowing_date}\nharvest date: {harvest_date}\ncustom cultivation period: {custom_harvest}"
        )

        def development_base_year(time_index, sowing_date, t_air, t_base):
            """
            Cumulative temperature experienced by plant as measure for plant development
            from sowing_date until end of the same year (base year)
            """

            delta_tt_list = []  # creating a list
            for temp, date in zip(t_air, time_index):
                if date < sowing_date:
                    delta_tt = 0
                else:
                    if temp > t_base:
                        delta_tt = (
                            temp - t_base
                        ) / 24  # SIMPLE crop model has daily temp values, convert to hourly
                    else:
                        delta_tt = 0
                delta_tt_list.append(delta_tt)
            cumulative_temp = np.cumsum(delta_tt_list)
            return cumulative_temp

        def development_extension(time_index, sowing_date, harvest_date, t_air, t_base):
            """
            Additional cumulative temperature experienced by plant as measure for plant development
            if growth extends to following year (until harvest_date if such is given).
            Note that harvest_date (MM-DD) has to be before sowing_date (MM-DD), growth cannot extend beyond one year in total.
            """
            delta_tt_list = []  # creating a list
            for temp, date in zip(t_air, time_index):
                if date < harvest_date < sowing_date:
                    if temp > t_base:
                        delta_tt = (
                            temp - t_base
                        ) / 24  # SIMPLE crop model has daily temp values, convert to hourly
                    else:
                        delta_tt = 0
                else:
                    delta_tt = 0
                delta_tt_list.append(delta_tt)
            cumulative_temp = np.cumsum(delta_tt_list)
            return cumulative_temp

        def development_cache(
            time_index,
            sowing_date,
            harvest_date,
            cum_temp_base_cache,
            cum_temp_ext_cache,
        ):
            """
            Cumulative temperature experienced in the base year cached for extension in the following year,
            cumulative temperature experienced in the following year until harvest_date removed afterwards
            so it does not interfer with base year [K]
            """
            delta_tt_list = []  # creating a list
            for date in time_index:
                if date <= harvest_date < sowing_date:
                    delta_tt = cum_temp_base_cache
                else:
                    delta_tt = -cum_temp_ext_cache
                delta_tt_list.append(delta_tt)
            return np.array(delta_tt_list)

        # Create 3 lists for cumulative_temperature: base year, following year (extension), cached values
        cumulative_temp_1 = development_base_year(dates, sowing_date, t_air, t_base)
        cum_temp_base_cache = cumulative_temp_1[-1]

        cumulative_temp_2 = development_extension(
            dates, sowing_date, harvest_date, t_air, t_base
        )
        cum_temp_ext_cache = cumulative_temp_2[-1]

        cumulative_temp_3 = development_cache(
            dates, sowing_date, harvest_date, cum_temp_base_cache, cum_temp_ext_cache
        )

        # Add the three lists together to get total cumulative temperature
        cumulative_temp = cumulative_temp_1 + cumulative_temp_2 + cumulative_temp_3

        # Update t_sum if custom_harvest = True (custom harvest date provided instead of maturity according to SIMPLE)
        t_sum = cumulative_temp[dates.index(harvest_date)] if custom_harvest else t_sum
        # print(f"t_sum: {t_sum}")

        # Assumption for transition point from growth period to senescence (decline) period: f_solar = 0.999f_solar_max
        # with f_solar being a function of cum_temp
        cum_temp_to_reach_f_solar_max = i50a - 100 * np.log(1 / 999)

        # f_solar(cum_temp) according to SIMPLE
        f_solar_list = []
        for cum_temp in cumulative_temp:
            if cum_temp < 1:
                f_solar = 0
            elif cum_temp < cum_temp_to_reach_f_solar_max:
                f_solar = f_solar_max / (1 + np.exp(-0.01 * (cum_temp - i50a)))
            elif cum_temp < t_sum:
                f_solar = f_solar_max / (1 + np.exp(0.01 * (cum_temp - (t_sum - i50b))))
            else:
                f_solar = 0
            f_solar_list.append(f_solar)

        return np.array(f_solar_list)

    def calc_Ftemp(self, t_air, t_opt, t_base, **kwargs):
        r"""
        Calculates the temperature on the biomass rate
        ----
        Parameters
        ----------
        t_air: ambient temperature as pd.series or list
        t_opt: optimum temperature for biomass growth
        t_base: base temperature for biomass growth

        Returns
        -------
        te : list of numerical values:
             temperature coefficients for calculating biomass rate

        Notes
        -----
        Corresponds to Fig 1.(a) of https://doi.org/10.1016/j.eja.2019.01.009


        """

        # Check if input arguments have proper type and length
        if not isinstance(t_air, (list, pd.Series)):
            print("Argument 'temp' is not of type list or pd.Series!")
        te = []  # creating a list
        # Calculate te
        for t in t_air:
            if t < t_base:
                x = 0
                te.append(x)

            elif t_base <= t <= t_opt:
                x = (t - t_base) / (t_opt - t_base)
                te.append(x)

            elif t > t_opt:
                x = 1
                te.append(x)

        return np.array(te)

    def calc_Fwater(self, et_o, vwc, s_water, **kwargs):
        r"""
        Calculates the soil water availability impact on the biomass rate
        arid factor derived from simple crop model and Woli et al. (2012), divided by 24 to transform to hourly values
        ----
        Parameters
        ----------
        et_o: potential evapotranspiration
        vwc: volumetric water content
        s_water: Sensitivity of RUE (or harvest index) to drought stress (ARID index)

        Returns
        -------
        arid : list of numerical values:
             aridity factor affecting the biomass rate

        Notes
        -----
        Corresponds to Fig 1.(d) of https://doi.org/10.1016/j.eja.2019.01.009


        """

        def arid(et_o, vwc):
            et_o = np.array(et_o)
            vwc = np.array(vwc)

            # Calculate arid
            wi = 1 - np.minimum(abs(et_o), 0.096 * vwc) / abs(et_o)
            return wi

        f_water_list = 1 - s_water * arid(et_o, vwc)
        return np.array(f_water_list)

    def calc_Fheat(self, t_air, t_max, t_ext, **kwargs):
        r"""
        Calculates the heat impact on the biomass rate
        ----
        Parameters
        ----------
        t_air: Hourly timeseries of air temperature as pd.series or list
        t_max: Threshold temperature to start accelerating senescence from heat stress (°C)
        t_ext: Threshold temperature when biomass growth rate starts to be reduced by heat stress (°C)
        Returns
        -------
        hi : list of numerical values:
             temperature coefficients for calculating biomass rate

        Notes
        -----
        Corresponds to Fig 1.(b) of https://doi.org/10.1016/j.eja.2019.01.009

        """

        def paper_function(t_max, t_heat, t_ext):
            """function witin https://doi.org/10.1016/j.eja.2019.01.009"""
            if t_max <= t_heat:
                x = 1
            elif t_heat < t_max <= t_ext:
                x = 1 - (t_max - t_heat) / (t_ext - t_heat)
            elif t_max > t_ext:
                x = 0
            return x

        # Check if input arguments have proper type and length
        if not isinstance(t_air, (list, pd.Series, np.ndarray)):
            raise TypeError("Argument t_air is not of type list or pd.Series!")

        n_timesteps = len(t_air)

        if n_timesteps == 0:
            raise ValueError("Argument t_air is empty")

        n_hours_in_day = 24

        n_days, n_hours = np.divmod(n_timesteps, n_hours_in_day)

        # in the paper they provide T_heat as T_max in the table 1a
        t_heat = t_max
        hi = []  # creating a list

        if n_days == 0:
            # there is less than a day, returing same value n_timesteps times
            n_timesteps = n_hours
            t_max = np.max(t_air)
            hi.append(np.ones(n_timesteps) * paper_function(t_max, t_heat, t_ext))
        else:
            # there is one day or more
            if n_hours != 0:
                N = n_days + 1
            else:
                N = n_days
            for i in range(0, N):
                if i == n_days:
                    if n_hours != 0:
                        # last day is not a full day
                        n_timesteps = n_hours
                    else:
                        # last day is a full day
                        n_timesteps = n_hours_in_day
                else:
                    # it is a full day
                    n_timesteps = n_hours_in_day
                t_max = np.max(
                    t_air[(i * n_hours_in_day) : (i * n_hours_in_day + n_timesteps)]
                )
                hi.append(np.ones(n_timesteps) * paper_function(t_max, t_heat, t_ext))

        return np.hstack(hi)

    @property
    def efficiency(self):

        # Conversion factor Watt Hours (WH) to Mega Joules (MJ)
        c_wh_to_mj = 3.6e-3
        c_kg_to_g = 1e-3
        crop_params = crop_dict[self.crop_type]
        fTEMP = self.calc_Ftemp(self.t_air, **crop_params)
        fWATER = self.calc_Fwater(self.et_0, self.vwc, **crop_params)
        fHEAT = self.calc_Fheat(t_air=self.t_air, **crop_params)
        fSOLAR = self.calc_Fsolar(
            time_index=self.time_index,
            t_air=self.t_air,
            sowing_date=self.sowing_date,
            harvest_date=self.harvest_date,
            **crop_params,
        )

        rue = self.get_crop_param("rue")

        return rue * fSOLAR * fTEMP * np.minimum(fWATER, fHEAT) * c_wh_to_mj * c_kg_to_g

    def build_solph_components(self):
        """ """
        hi = self.get_crop_param("hi")
        self.conversion_factors.update(
            {
                self.solar_bus: sequence(1),
                self.harvest_bus: sequence(hi * abs(self.efficiency)),
            }
        )

        self.inputs.update(
            {
                self.solar_bus: Flow(
                    variable_costs=self.carrier_cost, **self.input_parameters
                )
            }
        )

        self.outputs.update(
            {
                self.harvest_bus: Flow(
                    nominal_value=self._nominal_value(),
                    variable_costs=self.marginal_cost,
                    investment=self._investment(),
                    **self.output_parameters,
                )
            }
        )

        if self.biomass_bus is not None:
            self.conversion_factors.update(
                {self.biomass_bus: sequence((1 - hi) * abs(self.efficiency))}
            )
            self.outputs.update(
                {
                    self.biomass_bus: Flow(),
                }
            )


@dataclasses.dataclass(unsafe_hash=False, frozen=False, eq=False)
class MimoCrop(MIMO):
    r"""
    Crop Model Converter with 3 inputs (irradiation, precipitation, irrigation)
    and 2 outputs (crop harvest, remaining biomass), based on MultipleInputMultipleOutputConverter.

    Functions designed according to 3 publications, please note the abbreviations in the function descriptions:
    FAO56       ISBN: 978-92-5-104219-9
    SIMPLE      https://doi.org/10.1016/j.eja.2019.01.009
    ARID        https://doi.org/10.2134/agronj2011.0286

    Parameters
    ----------
    type: str
        will link MimoCrop to TYPEMAP, currently named 'mimo-crop'
    name: str
        a remarkable name for your component
    tech: str
        whatever, I just specified it as 'mimo'
    carrier: str
        carrier of the primary bus, should be overwritten by carriers directly assigned to buses
    primary: str
        primary bus object to which investments etc. are assigned
    expandable: boolean or numeric (binary)
        True, if capacity can be expanded within optimization. Default: False.
    capacity: numeric
        The capacity of crop. It is expressed in cultivated area [m²]
    capacity_minimum: numeric
        Minimum invest capacity in unit of output capacity.
    capacity_potential: numeric   (seems not to work as it should)
        Maximum invest capacity in unit of output capacity.
    capacity_cost: numeric
        Investment costs per unit of output capacity. If capacity is not set,
        this value will be used for optimizing the conversion output capacity.
    crop_type: str
        the name of crop as defined in global_specs/crop_specs.py
    solar_energy_bus: oemof.solph.Bus
        An oemof bus instance where the MimoCrop unit is connected to with
        its input, it is expected to provide W/m² irradiance.
    precipitation_bus: oemof.solph.Bus
        An oemof bus instance where the MimoCrop unit is connected to with
        its input, it is expected to provide mm precipitation.
    irrigation_bus: oemof.solph.Bus
        An oemof bus instance where the MimoCrop unit is connected to with
        its input, it is expected to provide mm irrigation.
    crop_bus: oemof.solph.Bus
        An oemof bus instance where the MimoCrop unit is connected to with
        its crop-harvest output. The unit is kg.
    biomass_bus: oemof.solph.Bus
        An oemof bus instance where the MimoCrop is connected with its
        non-edible, remaining (residual) biomass output. The unit is kg.
    time_profile: time series
        time profile separate from time index
    ghi_profile: time series
        global horizontal solar irradiance in W/m²
    tp_profile: time series
        total precipitation in mm
    t_air_profile: time series
        ambient air tempearture in °C
    t_dp_profile: time series
        dew point temperature in °C
    windspeed_profile: time series
        windspeed in m/s
    elevation: float
        elevation above sea level of the location
    crop_type: str
        the name of crop as defined in global_specs/crop_specs.py
    sowing_date: str
        date when cultivation period starts, MM-DD format
    harvest_date:
        date when cultivation period ends, MM-DD format
    has_irrigation: bool
        if set to False, irrigation profile will be set to 0

    Other optional parameters that could be added:
    ---------------------------------------------
    marginal_cost: numeric
        Marginal cost for one unit of produced output. Default: 0
    carrier_cost: numeric
        Carrier cost for one unit of used input. Default: 0
    input_parameters: dict (optional)
        Set parameters on the input edge of the conversion unit
        (see oemof.solph for more information on possible parameters)
    output_parameters: dict (optional)
        Set parameters on the output edge of the conversion unit
        (see oemof.solph for more information on possible parameters)

    Notes:
    ---------------------------------------------
    One input (eg. ghi) has to be volatile with a fixed capacity, other 2 dispatchable.
    MimoCrop capacity_potential seems to not work correctly.
    Two options to successfully run MimoCrop:
    1) Expandable = True and ignore capacity, use capacity of volatile source instead.
    2) Expandable = False, set capacity equal to that of the volatile source.

    In case you experience a weird error, for example "TypeError: can't multiply sequence by non-int of type 'float'",
    this is likely due to empty rows at the end of your input csv files.
    Please remove them and try again.
    """

    type: str = "mimo-crop"

    name: str = ""

    tech: str = "mimo"

    carrier: str = ""

    primary: str = ""

    expandable: bool = False

    capacity: float = 0

    capacity_minimum: float = None

    capacity_potential: float = None

    capacity_cost: float = 0

    solar_energy_bus: Bus = None

    precipitation_bus: Bus = None

    irrigation_bus: Bus = None

    crop_bus: Bus = None

    biomass_bus: Bus = None

    time_profile: Union[float, Sequence[float]] = None

    ghi_profile: Union[float, Sequence[float]] = None

    tp_profile: Union[float, Sequence[float]] = None

    t_air_profile: Union[float, Sequence[float]] = None

    t_dp_profile: Union[float, Sequence[float]] = None

    windspeed_profile: Union[float, Sequence[float]] = None

    elevation: float = 0

    crop_type: str = ""  # according to crop_dict

    sowing_date: str = ""  # MM-DD format

    harvest_date: str = ""  # MM-DD format

    has_irrigation: bool = False

    def __init__(self, **kwargs):
        """
        Preprocessing of crop input data to calculate conversion_factors
        and assign these together with the correct busses to MIMO parent class
        """

        def specify_cultivation_parameters(dates, sowing_date, harvest_date):
            """
            Convert all time-based elements to Timestamp objects,
            adapt sowing_date and harvest_date to input time series (year and timezone),
            specify sowing_date and harvest_date if not provided
            """
            year = str(dates[0].year)
            timezone = dates[0].tz
            # Opt. 1: sowing_date and harvest date given, plant maturity (t_sum) will be updated (custom_harvest=True)
            if sowing_date and harvest_date:
                sowing_date = pd.Timestamp(year + "-" + sowing_date).tz_localize(
                    timezone
                )
                harvest_date = pd.Timestamp(year + "-" + harvest_date).tz_localize(
                    timezone
                )
                has_custom_harvest = True
                # If harvest and sowing date are the same, move harvest date one time step back to avoid problems
                if sowing_date == harvest_date:
                    harvest_date = dates[dates.index(harvest_date) - 1]
            # Opt. 2: only sowing date, harvest_date (end of cultivation period) is one day before (following year),
            # maturity according to SIMPLE
            elif sowing_date and not harvest_date:
                sowing_date = pd.Timestamp(year + "-" + sowing_date).tz_localize(
                    timezone
                )
                harvest_date = dates[dates.index(sowing_date) - 1]
                has_custom_harvest = False
            # Opt. 3: no dates, growth from start of the year till maturity (from SIMPLE)
            else:
                sowing_date = dates[0]
                harvest_date = dates[-1]
                has_custom_harvest = False

            return {
                "sowing_date": sowing_date,
                "harvest_date": harvest_date,
                "has_custom_harvest": has_custom_harvest,
            }

        def development_base_year(date, t_air, sowing_date, t_base, **kwargs):
            """
            Cumulative temperature experienced by plant as measure for plant development
            from sowing_date until end of the same year (base year) (SIMPLE) [K]
            """
            if date < sowing_date:
                delta_tt = 0
            else:
                if t_air > t_base:
                    delta_tt = (
                        t_air - t_base
                    ) / 24  # SIMPLE crop model has daily temp values, convert to hourly
                else:
                    delta_tt = 0
            return delta_tt

        def development_extension(
            date, t_air, sowing_date, harvest_date, t_base, **kwargs
        ):
            """
            Additional cumulative temperature experienced by plant as measure for plant development
            if growth extends to following year
            (harvest_date < sowing_date in MM-DD format because it would be in the next year) (SIMPLE) [K]
            """
            if date < harvest_date < sowing_date:
                if t_air > t_base:
                    delta_tt = (
                        t_air - t_base
                    ) / 24  # SIMPLE crop model has daily temp values, convert to hourly
                else:
                    delta_tt = 0
            else:
                delta_tt = 0
            return delta_tt

        def development_cache(
            date,
            cum_temp_base_cache,
            cum_temp_ext_cache,
            harvest_date,
            sowing_date,
            **kwargs,
        ):
            """
            Cumulative temperature experienced in the base year cached for extension in the following year,
            cumulative temperature experienced in the following year until harvest_date removed from base year
            (SIMPLE) [K]
            """
            if date <= harvest_date < sowing_date:
                delta_tt = cum_temp_base_cache
            else:
                delta_tt = -cum_temp_ext_cache
            return delta_tt

        def custom_cultivation_period(
            df, harvest_date, has_custom_harvest, t_sum, **kwargs
        ):
            """Calculates new t_sum for plant growth curve if custom harvest_date is given [K]"""
            if has_custom_harvest and harvest_date in df.index:
                return {"t_sum": df.loc[harvest_date, "cum_temp"]}
            else:
                return {"t_sum": t_sum}

        def temp(t_air, t_opt, t_base, **kwargs):
            """Temperature effect on plant growth (SIMPLE) [-]"""
            if t_air < t_base:
                f_temp = 0
            elif t_base <= t_air < t_opt:
                f_temp = (t_air - t_base) / (t_opt - t_base)
            else:
                f_temp = 1
            return f_temp

        def heat(t_air, t_max, t_ext, **kwargs):
            """Heat stress effect on plant growth (SIMPLE) [-]"""
            if t_air <= t_max:
                f_heat = 1
            elif t_max < t_air <= t_ext:
                f_heat = 1 - (t_air - t_max) / (t_ext - t_max)
            else:
                f_heat = 0
            return f_heat

        def soil_heat_flux(ghi, irr_w, **kwargs):
            """Soil heat flux as fraction of incoming radiation (FAO56) [W/m²]"""
            if ghi > 1e-5:
                g = 0.1 * irr_w
            else:
                g = 0.5 * irr_w
            return g

        def potential_evapotranspiration(z, t_air, t_dp, w10, irr_w, g_w, **kwargs):
            """Potential evapotranspiration for reference grass (FAO56) [mm]"""
            cp_air = 1.013e-3  # specific heat at constant pressure [MJ/(kg °C)]
            epsilon = 0.622  # ratio molecular weight of water vapour/dry air [-]
            h_vap = 2.45  # latent heat of vaporization [MJ/kg]
            rho_w = 1000  # density of water [kg/m³]
            k_c = 1  # crop coefficient (FAO) [-]
            irr = 86.4e-3 * irr_w  # W/m² to MJ/m²*day
            g = 3.6e-3 * g_w  # W/m² to MJ/m²*day
            w2 = (
                w10 * 4.87 / np.log(672.58)
            )  # wind speed at 2m above ground (FAO56) [m/s]
            p = 101.3 * ((293 - 0.0065 * z) / 293) ** 5.26  # atmospheric pressure [kPa]
            gamma = cp_air * p / (h_vap * epsilon)  # psychrometric constant
            delta = (
                4098
                * (0.6108 * np.exp(17.27 * t_air / (t_air + 237.3)))
                / (t_air + 237.3) ** 2
            )  # slope of sat vap press curve

            def vap_pressure(t):
                """Water vapor saturation pressure at specific temperature [kPa]"""
                e = 0.6108 * np.exp(17.27 * t / (t + 237.3))
                return e

            e_s = vap_pressure(t_air)  # saturation vapour pressure [kPa]
            e_a = vap_pressure(
                t_dp
            )  # actual vapour pressure (sat vap press at dewpoint temp) [kPa]

            et_0 = (
                0.408 * delta * (irr - g)
                + gamma * 900 / (t_air + 273) * w2 * (e_s - e_a)
            ) / (
                delta + gamma * 1.34 * w2
            )  # [mm/m²*day]
            et_p = k_c * et_0 / 24  # [mm/m²*h]
            # q_et = et * rho_w * h_vap / 1000  # [W/m²]
            return et_p

        def runoff(p, rcn, **kwargs):
            """Surface runoff of precipitation (ARID) [mm]"""
            s = 25400 / rcn - 254
            i_a = 0.2 * s
            if p > i_a:
                r = (p - i_a) ** 2 / (p + i_a - s) / 24
            else:
                r = 0.0
            return r

        def soil_water_balance(df, has_irrigation, awc, ddc, rzd, **kwargs):
            """
            Soil water content (swc) [-] and irrigation [mm] (if True) for timestep i based on
            precipitation (tp) [mm], surface runoff [mm] for timestep i as well as
            soil water content, deep drainage [mm] and actual evapotranspiration (et_a) [mm] for timestep i-1
            (ARID)
            """
            df["deep_drain"] = 0.0
            df["et_a"] = 0.0
            df["irrigation"] = 0.0
            df["swc"] = 0.0

            def deep_drainage(ddc, rzd, water_cap, water_con_bd, **kwargs):
                """Deep drainage of soil water (ARID) [mm]"""
                if water_con_bd > water_cap:
                    d = ddc * rzd * (water_con_bd - water_cap) / 24
                else:
                    d = 0
                return d

            swc_cache = awc
            df.loc[df.index[0], "swc"] = swc_cache
            for index, row in df.iloc[1:].iterrows():
                df.loc[index, "deep_drain"] = deep_drainage(
                    ddc=ddc, rzd=rzd, water_cap=awc, water_con_bd=swc_cache
                )
                df.loc[index, "et_a"] = min(
                    0.096 * rzd * swc_cache / 24, row["et_p"]
                )  # 0.096 water uptake constant (SIMPLE)

                if has_irrigation:
                    water_deficit = df.loc[index, "et_a"] - (row["tp"] - row["runoff"])
                    df.at[index, "irrigation"] = max(water_deficit, 0)

                delta_swc = (
                    row["tp"]
                    - row["runoff"]
                    - df.loc[index, "deep_drain"]
                    - df.loc[index, "et_a"]
                    + df.loc[index, "irrigation"]
                ) / rzd

                df.loc[index, "swc"] = swc_cache + delta_swc
                swc_cache = df.loc[index, "swc"]

        def drought(et_p, et_a, s_water, **kwargs):
            """Drought stress effect on plant growth (SIMPLE) [-]"""
            arid = 1 - min(et_p, et_a) / et_p
            f_drought = 1 - s_water * arid
            return f_drought

        def faster_senescence(cum_temp, f_heat, f_drought, i50maxh, i50maxw, **kwargs):
            """Heat and drought stress effect on solar interception due to faster canopy senescence (SIMPLE) [-]"""
            if cum_temp < 1:
                delta_i50b = 0
            else:
                delta_i50b = (i50maxh * (1 - f_heat) + i50maxw * (1 - f_drought)) / 24
            return delta_i50b

        def solar_interception(
            cum_temp, delta_i50b, t_sum, i50a, i50b, f_solar_max, **kwargs
        ):
            """
            Interception of incoming radiation by plant canopy
            according to cumulative temperature experienced (SIMPLE) [-]
            """
            cum_temp_to_reach_f_solar_max = i50a - 100 * np.log(
                1 / 999
            )  # derived from f_solar = 0.999f_solar_max
            if cum_temp < 1:
                f_solar = 0
            elif cum_temp < cum_temp_to_reach_f_solar_max:
                i50b += delta_i50b
                f_solar = f_solar_max / (1 + np.exp(-0.01 * (cum_temp - i50a)))
            elif cum_temp < t_sum:
                i50b += delta_i50b
                f_solar = f_solar_max / (1 + np.exp(0.01 * (cum_temp - (t_sum - i50b))))
            else:
                f_solar = 0
            return f_solar

        # Get attributes for biomass modelling from kwargs
        self.elevation = kwargs.pop("elevation")
        self.crop_type = kwargs.pop("crop_type")
        self.sowing_date = kwargs.pop("sowing_date")
        self.harvest_date = kwargs.pop("harvest_date")
        self.has_irrigation = kwargs.pop("has_irrigation")

        # Create DataFrame out of input profiles (time series), time_profile will be set as DatetimeIndex
        profiles_dict = {
            key.replace("_profile", ""): value
            for key, value in kwargs.items()
            if key.endswith("profile")
        }
        time_index = profiles_dict.pop("time")
        profiles_df = pd.DataFrame(data=profiles_dict, index=time_index)

        # Get crop and soil parameters from database, calculate cultivation parameters
        crop_params = crop_dict[self.crop_type]
        soil_params = soil_dict[self.crop_type]
        cultivation_params = specify_cultivation_parameters(
            profiles_df.index, self.sowing_date, self.harvest_date
        )

        # Apply functions for biomass modelling on DataFrame
        profiles_df["cum_temp_base_year"] = profiles_df.apply(
            lambda row: development_base_year(
                row.name, row["t_air"], **crop_params, **cultivation_params
            ),
            axis=1,
        ).cumsum()

        profiles_df["cum_temp_extension"] = profiles_df.apply(
            lambda row: development_extension(
                row.name, row["t_air"], **crop_params, **cultivation_params
            ),
            axis=1,
        ).cumsum()

        cum_temp_base_cache = profiles_df["cum_temp_base_year"].iat[-1]
        cum_temp_ext_cache = profiles_df["cum_temp_extension"].iat[-1]

        profiles_df["cum_temp_cache"] = profiles_df.apply(
            lambda row: development_cache(
                row.name, cum_temp_base_cache, cum_temp_ext_cache, **cultivation_params
            ),
            axis=1,
        )

        profiles_df["cum_temp"] = (
            profiles_df["cum_temp_base_year"]
            + profiles_df["cum_temp_extension"]
            + profiles_df["cum_temp_cache"]
        )

        crop_params.update(
            custom_cultivation_period(profiles_df, **cultivation_params, **crop_params)
        )

        profiles_df["f_temp"] = profiles_df["t_air"].apply(
            lambda t_air: temp(t_air, **crop_params)
        )

        profiles_df["f_heat"] = profiles_df["t_air"].apply(
            lambda t_air: heat(t_air, **crop_params)
        )

        profiles_df["rad_net"] = (
            profiles_df["ghi"] * 0.67
        )  # assuming ground albedo of 0.23 for g and et_p (FAO56)

        profiles_df["g"] = profiles_df.apply(
            lambda row: soil_heat_flux(row["ghi"], row["rad_net"]), axis=1
        )

        profiles_df["et_p"] = profiles_df.apply(
            lambda row: potential_evapotranspiration(
                self.elevation,
                row["t_air"],
                row["t_dp"],
                row["windspeed"],
                row["rad_net"],
                row["g"],
            ),
            axis=1,
        )

        profiles_df["runoff"] = profiles_df["tp"].apply(
            lambda tp: runoff(tp, **soil_params)
        )

        soil_water_balance(profiles_df, self.has_irrigation, **soil_params)

        profiles_df["f_drought"] = profiles_df.apply(
            lambda row: drought(row["et_p"], row["et_a"], **crop_params), axis=1
        )

        profiles_df["delta_i50b"] = profiles_df.apply(
            lambda row: faster_senescence(
                row["cum_temp"], row["f_heat"], row["f_drought"], **crop_params
            ),
            axis=1,
        ).cumsum()

        profiles_df["f_solar"] = profiles_df.apply(
            lambda row: solar_interception(
                row["cum_temp"], row["delta_i50b"], **crop_params
            ),
            axis=1,
        )

        # Set irrigation to 0 outside the cultivation period (f_solar == 0)
        profiles_df.loc[profiles_df["f_solar"] == 0, "irrigation"] = 0

        # Calculate total biomass rate
        profiles_df["total_biomass"] = (
            crop_params["rue"]
            * profiles_df["ghi"]
            * 3.6e-3  # Wh to MJ
            * profiles_df["f_solar"]
            * profiles_df["f_temp"]
            * profiles_df[["f_heat", "f_drought"]].min(axis=1)
            * 1e-3  # g to kg
        )

        # Split into crop harvest and remaining (residual) biomass
        profiles_df["crop_harvest"] = crop_params["hi"] * profiles_df["total_biomass"]
        profiles_df["remaining_biomass"] = (
            profiles_df["total_biomass"] - profiles_df["crop_harvest"]
        )

        # Replace all values smaller than 1e-10 with 1e-10 (incl. negative values), MIMO does not handle values '0' well
        profiles_df = profiles_df.mask((profiles_df <= 1e-5), 1e-10)

        # round all floats to 10 decimal places
        profiles_df = profiles_df.round(10)

        # Update conversion factors for the different inputs and outputs
        conversion_factors = {}
        for key, value in kwargs.items():
            if hasattr(value, "type") and value.type == "bus":
                if key.startswith("solar_energy_bus"):
                    conversion_factors[f"conversion_factor_{value}"] = sequence(1)
                if key.startswith("precipitation_bus"):
                    conversion_factors[f"conversion_factor_{value}"] = sequence(
                        profiles_df["tp"] / profiles_df["ghi"]  # input data
                    )
                if key.startswith("irrigation_bus"):
                    conversion_factors[f"conversion_factor_{value}"] = sequence(
                        profiles_df["irrigation"]
                        / profiles_df["ghi"]  # pre-processed input data
                    )
                if key.startswith("crop_bus"):
                    conversion_factors[f"conversion_factor_{value}"] = sequence(
                        profiles_df["crop_harvest"]
                        / profiles_df["ghi"]  # pre-processed input data
                    )
                if key.startswith("biomass_bus"):
                    conversion_factors[f"conversion_factor_{value}"] = sequence(
                        profiles_df["remaining_biomass"]
                        / profiles_df["ghi"]  # pre-processed input data
                    )
        kwargs.update(conversion_factors)

        self.solar_energy_bus = kwargs.pop("solar_energy_bus")
        self.precipitation_bus = kwargs.pop("precipitation_bus")
        self.irrigation_bus = kwargs.pop("irrigation_bus")
        self.crop_bus = kwargs.pop("crop_bus")
        self.biomass_bus = kwargs.pop("biomass_bus")

        # Initializes MIMO with the crop-specific buses and conversion_factors
        super().__init__(
            from_bus_0=self.solar_energy_bus,
            from_bus_1=self.precipitation_bus,
            from_bus_2=self.irrigation_bus,
            to_bus_0=self.crop_bus,
            to_bus_1=self.biomass_bus,
            **kwargs,
        )

        # Other mandatory arguments
        self.type = kwargs.pop("type", None)
        self.name = kwargs.pop("name", None)
        self.carrier = kwargs.pop("carrier", None)
        self.tech = kwargs.pop("tech", None)
        self.primary = kwargs.pop("primary", None)
        self.expandable = kwargs.pop("expandable", None)
        self.capacity = kwargs.pop("capacity", None)
        self.capacity_minimum = kwargs.pop("capacity_minimum", None)
        self.capacity_potential = kwargs.pop("capacity_potential", None)
        self.capacity_cost = kwargs.pop("capacity_cost", None)



from dataclasses import field
from typing import Sequence, Union

from oemof.solph._plumbing import sequence
from oemof.solph.buses import Bus
from oemof.solph.components import Converter
from oemof.solph.flows import Flow

from oemof.tabular._facade import dataclass_facade, Facade


@dataclass_facade
class WaterFiltration(Converter, Facade):
    r"""WaterFiltration unit with two inputs and one output.

    Parameters
    ----------
    electricity_bus: oemof.solph.Bus
        An oemof bus instance where unit is connected to with
        its low temperature input.
    water_in_bus: oemof.solph.Bus
        An oemof bus instance where unit is connected to with
        its water input.
    water_out_bus: oemof.solph.Bus
        An oemof bus instance where the unit is connected to with
        its water output.
    capacity: numeric
        The thermal capacity (high temperature output side) of the unit.
    carrier_cost: numeric
        Carrier cost for one unit of used input. Default: 0
    capacity_cost: numeric
        Investment costs per unit of output capacity.
        If capacity is not set, this value will be used for optimizing the
        conversion output capacity.
    expandable: boolean or numeric (binary)
        True, if capacity can be expanded within optimization. Default: False.
    lifetime: int (optional)
        Lifetime of the component in years. Necessary for multi-period
        investment optimization.
        Note: Only applicable for a multi-period model. Default: None.
    age : int (optional)
        The initial age of a flow (usually given in years);
        once it reaches its lifetime (considering also
        an initial age), the flow is forced to 0.
        Note: Only applicable for a multi-period model. Default: 0.
    fixed_costs : numeric (iterable or scalar) (optional)
        The fixed costs associated with a flow.
        Note: Only applicable for a multi-period model. Default: None.
    capacity_potential: numeric
        Maximum invest capacity in unit of output capacity. Default: +inf.
    input_parameters: dict (optional)
        Set parameters on the input edge of the conversion unit
         (see oemof.solph for more information on possible parameters)
    output_parameters: dict (optional)
        Set parameters on the output edge of the conversion unit
         (see oemof.solph for more information on possible parameters)
    """

    electricity_bus: Bus

    water_in_bus: Bus

    water_out_bus: Bus

    tech: str

    carrier: str = ""

    capacity: float = None

    marginal_cost: float = 0

    carrier_cost: float = 0

    capacity_cost: float = None

    expandable: bool = False

    lifetime: int = None

    age: int = 0

    fixed_costs: Union[float, Sequence[float]] = None

    capacity_potential: float = float("+inf")

    input_parameters: dict = field(default_factory=dict)

    output_parameters: dict = field(default_factory=dict)

    def build_solph_components(self):
        """TODO change efficiencies here"""
        self.conversion_factors.update(
            {
                self.electricity_bus: sequence(1),
                self.water_in_bus: sequence(1),
                self.water_out_bus: sequence(1),
            }
        )

        self.inputs.update(
            {
                self.electricity_bus: Flow(
                    variable_costs=self.carrier_cost, **self.input_parameters
                ),
                self.water_in_bus: Flow(),
            }
        )

        self.outputs.update(
            {
                self.water_out_bus: Flow(
                    nominal_value=self._nominal_value(),
                    variable_costs=self.marginal_cost,
                    investment=self._investment(),
                )
            }
        )


@dataclass_facade
class WaterPump(Converter, Facade):
    r"""WaterPump unit with two inputs and one output.

    Parameters
    ----------
    electricity_bus: oemof.solph.Bus
        An oemof bus instance where unit is connected to with
        its low temperature input.
    water_in_bus: oemof.solph.Bus
        An oemof bus instance where unit is connected to with
        its water input.
    water_out_bus: oemof.solph.Bus
        An oemof bus instance where the unit is connected to with
        its water output.
    capacity: numeric
        The thermal capacity (high temperature output side) of the unit.
    carrier_cost: numeric
        Carrier cost for one unit of used input. Default: 0
    capacity_cost: numeric
        Investment costs per unit of output capacity.
        If capacity is not set, this value will be used for optimizing the
        conversion output capacity.
    expandable: boolean or numeric (binary)
        True, if capacity can be expanded within optimization. Default: False.
    lifetime: int (optional)
        Lifetime of the component in years. Necessary for multi-period
        investment optimization.
        Note: Only applicable for a multi-period model. Default: None.
    age : int (optional)
        The initial age of a flow (usually given in years);
        once it reaches its lifetime (considering also
        an initial age), the flow is forced to 0.
        Note: Only applicable for a multi-period model. Default: 0.
    fixed_costs : numeric (iterable or scalar) (optional)
        The fixed costs associated with a flow.
        Note: Only applicable for a multi-period model. Default: None.
    capacity_potential: numeric
        Maximum invest capacity in unit of output capacity. Default: +inf.
    input_parameters: dict (optional)
        Set parameters on the input edge of the conversion unit
         (see oemof.solph for more information on possible parameters)
    output_parameters: dict (optional)
        Set parameters on the output edge of the conversion unit
         (see oemof.solph for more information on possible parameters)
    """

    electricity_bus: Bus

    water_in_bus: Bus

    water_out_bus: Bus

    tech: str

    carrier: str = ""

    capacity: float = None

    marginal_cost: float = 0

    carrier_cost: float = 0

    capacity_cost: float = None

    expandable: bool = False

    lifetime: int = None

    age: int = 0

    fixed_costs: Union[float, Sequence[float]] = None

    capacity_potential: float = float("+inf")

    input_parameters: dict = field(default_factory=dict)

    output_parameters: dict = field(default_factory=dict)

    def build_solph_components(self):
        """TODO change efficiencies here"""
        self.conversion_factors.update(
            {
                self.electricity_bus: sequence(1),
                self.water_in_bus: sequence(1),
                self.water_out_bus: sequence(1),
            }
        )

        self.inputs.update(
            {
                self.electricity_bus: Flow(
                    variable_costs=self.carrier_cost, **self.input_parameters
                ),
                self.water_in_bus: Flow(),
            }
        )

        self.outputs.update(
            {
                self.water_out_bus: Flow(
                    nominal_value=self._nominal_value(),
                    variable_costs=self.marginal_cost,
                    investment=self._investment(),
                )
            }
        )



from dataclasses import dataclass
from oemof.tabular.constraint_facades import ConstraintFacade
from oemof_tabular_plugins.general.constraints.constraints import (
    renewable_share_minimum,
)


@dataclass
class MinimumRenewableShare(ConstraintFacade):
    name: str
    type: str
    limit: float
    keyword: str = "renewable_factor"

    def build_constraint(self, model):
        # to use the constraints in oemof.solph, we need to pass the model.
        # Check if there are flows with the keyword attribute
        flows = {}
        for i, o in model.flows:
            if hasattr(model.flows[i, o], self.keyword):
                flows[(i, o)] = model.flows[i, o]

        if not flows:
            raise Warning(f"No flows with keyword {self.keyword}")
        # else:
        # print(f"These flows will contribute to the integral limit: {flows.keys()}")

        # Add constraint to the model
        renewable_share_minimum(model, flows=flows, limit=self.limit)


CONSTRAINT_TYPE_MAP = {"minimum_renewable_share": MinimumRenewableShare}



"""
This file will contain all of the specific post-processing functionalities for water-energy-food-environment (WEFE)
systems e.g. calculating the crop yield and the total water footprint
"""


def calculate_crop_yield():
    return


def calculate_total_water_footprint():
    return



import os

import pandas as pd
from oemof.solph import EnergySystem, Model
from oemof.solph import processing
from oemof.solph.processing import parameter_as_dict

# TODO this should be with from oemof.tabular.datapackage import building when https://github.com/oemof/oemof-tabular/pull/173 is merged
from oemof_tabular_plugins.datapackage import building as otp_building

# ---- imports to be used when the package has been installed ----
from oemof.tabular import datapackage  # noqa
from oemof.tabular.facades import TYPEMAP

# ---- imports from oemof-tabular-plugins package ----
from oemof_tabular_plugins.general import (
    post_processing,
    CONSTRAINT_TYPE_MAP,
    pre_processing,
    logger,
)


def compute_scenario(
    scenario_dir,
    results_path,
    wacc,
    scenario_name=None,
    custom_attributes=None,
    typemap=None,
    moo=False,
    dash_app=False,
    parameters_units=None,
    infer_bus_carrier=True,
    skip_preprocessing=False,
    skip_infer_datapackage_metadata=False,
    save_raw_results=True,
):
    """

    Parameters
    ----------
    scenario_dir
    results_path
    wacc
    scenario_name
    custom_attributes
    typemap: default to oemof.tabular.facades.TYPEMAP
    moo
    skip_preprocessing: bool (opt)
        If True, the pre-processing to update input csv files based on cost parameters: CAPEX, OPEX fix, lifetime, WACC will not take place
        Default: False
    skip_infer_datapackage_metadata: bool (opt)
        If True, the file datapackage.json will not be updated with the possible changes in elements/*.csv and sequences/*.csv
        Default: False

    Returns
    -------

    """
    if scenario_name is None:
        scenario_name = os.path.basename(scenario_dir)

    if custom_attributes is None:
        custom_attributes = []

    if typemap is None:
        typemap = TYPEMAP

    # create results directory if it doesn't already exist
    if not os.path.exists(results_path):
        os.makedirs(results_path)

    if skip_preprocessing is False:
        # pre-processing to update input csv files based on cost parameters: CAPEX, OPEX fix, lifetime, WACC
        pre_processing(scenario_dir, wacc, custom_attributes, moo)

    if skip_infer_datapackage_metadata is False:
        otp_building.infer_metadata_from_data(
            package_name=scenario_name,
            path=scenario_dir,
            typemap=typemap,
        )

    # create energy system object from the datapackage
    es = EnergySystem.from_datapackage(
        os.path.join(scenario_dir, "datapackage.json"),
        attributemap={},
        typemap=typemap,
    )

    logger.info("Energy system created from datapackage")

    # create model from energy system (this is just oemof.solph)
    m = Model(es)
    logger.info("Model created from energy system")

    # add constraints from datapackage to the model
    m.add_constraints_from_datapackage(
        os.path.join(scenario_dir, "datapackage.json"),
        constraint_type_map=CONSTRAINT_TYPE_MAP,
    )
    logger.info("Constraints added to model")

    # if you want dual variables / shadow prices uncomment line below
    # m.receive_duals()

    # select solver 'gurobi', 'cplex', 'glpk' etc
    m.solve("cbc")

    # extract parameters and results
    params = parameter_as_dict(es)
    es.results = processing.results(m)
    if save_raw_results is True:
        es.dump(dpath=results_path, filename="oemof_raw")

    return post_processing(
        params,
        es,
        results_path,
        dp_path=os.path.join(scenario_dir, "datapackage.json"),
        dash_app=dash_app,
        parameters_units=parameters_units,
        infer_bus_carrier=infer_bus_carrier,
    )


def display_scenario_results(
    scenario_dir,
    results_path,
    dash_app=True,
    parameters_units=None,
):
    """

    Parameters
    ----------
    scenario_dir
    results_path
    Returns
    -------

    """

    # create results directory if it doesn't already exist
    if not os.path.exists(results_path):
        os.makedirs(results_path)

    oemof_dump_file = os.path.join(results_path, "oemof_raw")

    if os.path.exists(oemof_dump_file):
        es = EnergySystem()
        es.restore(dpath=results_path, filename="oemof_raw")

        params = None

        calculator = post_processing(
            params,
            es,
            results_path,
            dp_path=os.path.join(scenario_dir, "datapackage.json"),
            dash_app=dash_app,
            parameters_units=parameters_units,
            infer_bus_carrier=True,
        )
        return calculator
    else:
        raise FileNotFoundError(
            f"The file {oemof_dump_file} could not be found, you need to set the argument 'save_raw_results' to True when using the function 'compute_scenario'"
        )



import os
import shutil
import pytest
from oemof_tabular_plugins.script import compute_scenario
from oemof_tabular_plugins.datapackage.building import infer_metadata_from_data

from oemof_tabular_plugins.wefe import WEFE_TYPEMAP as TYPEMAP


class TestPostProcessingBusCarrierMapping:
    """
    This class contains tests for the possible ways to use (or not) the mapping of bus to carrier in the results dataframe
    """

    # store the relevant paths
    test_inputs_path = os.path.join("tests", "test_inputs")
    pre_p_dir = os.path.join(
        "tests",
        "test_inputs",
    )
    results_dir = os.path.join("tests", "tests_outputs")
    _package_path = os.path.join("data", "elements")
    package_path = _package_path

    def setup_method(self):
        """
        Sets up the testing environment before each test method is executed.
        """
        if os.path.exists(self.results_dir):
            shutil.rmtree(self.results_dir, ignore_errors=True)
        os.mkdir(self.results_dir)

    def compute_scenario(self, fname, infer_bus_carrier=True):
        """
        Run a single scenario and return a calculator instance
        """
        return compute_scenario(
            os.path.join(self.pre_p_dir, fname),
            self.results_dir,
            wacc=1,
            typemap=TYPEMAP,
            infer_bus_carrier=infer_bus_carrier,
            skip_preprocessing=True,
            skip_infer_datapackage_metadata=True,
        )

    def teardown_method(self):
        """
        Cleans up the testing environment after each test method is executed.
        """
        # removes the 'pre_processing' directory if it exists, ignoring errors
        if os.path.exists(self.results_dir):
            shutil.rmtree(self.results_dir, ignore_errors=True)

    def test_carrier_found_in_busses_logs_message(self, caplog):
        """
        The carriers are defined within a column 'carrier' of 'elements/bus.csv' file
        """
        f_name = "carriers_in_busses_only"
        infer_metadata_from_data(
            package_name=f_name,
            path=os.path.join(self.pre_p_dir, f_name),
            typemap=TYPEMAP,
        )
        # TODO add this assertion once oemof.tabular does not require carrier in its facade attributes (default to "")
        # calc = self.compute_scenario(f_name)
        # assert "carrier" in calc.df_results.index.names
        assert any(
            record.levelname == "INFO"
            and "Bus to carrier mapping found in 'elements/bus.csv' file of datapackage"
            in record.message
            for record in caplog.records
        )

    def test_carrier_found_in_busses_trumps_component_carriers(self, caplog):
        """
        If the carriers are defined within 'elements/bus.csv' file the carriers defined at the component level do not play a role
        """
        f_name = "carriers_over_defined"
        calc = self.compute_scenario(f_name)

        assert "carrier" in calc.df_results.index.names
        assert any(
            record.levelname == "INFO"
            and "Bus to carrier mapping found in 'elements/bus.csv' file of datapackage"
            in record.message
            for record in caplog.records
        )

    def test_carrier_inferred_from_components_carriers(self, caplog):
        """
        If the carriers are not defined within 'elements/bus.csv' file they will be inferred from the components
        """
        f_name = "carriers_in_components_only"
        calc = self.compute_scenario(f_name)

        assert "carrier" in calc.df_results.index.names
        assert any(
            record.levelname == "WARNING"
            and "the bus-carrier mapping will be inferred from the component's carrier"
            in record.message
            for record in caplog.records
        )

    def test_carrier_mapping_skipped(self, caplog):
        """
        If the carriers are not defined within 'elements/bus.csv' file and the user set infer_bus_carrier=False, then the results will simply not add a 'carrier' level in the MultiIndex of the results
        """
        f_name = "carriers_in_components_only"
        calc = self.compute_scenario(
            f_name,
            infer_bus_carrier=False,
        )
        assert "carrier" not in calc.df_results.index.names
        assert any(
            record.levelname == "INFO"
            and "Result dataframe will not contain 'carrier' in its MultiIndex levels."
            in record.message
            for record in caplog.records
        )

    def test_carrier_not_defined(self):
        """
        If the carriers are not defined within 'elements/bus.csv' file they will be inferred from the components, if the components themselves do not have carrier attributes an error is raised
        """
        f_name = "carriers_not_defined"
        with pytest.raises(ValueError) as e_info:
            infer_metadata_from_data(
                package_name=f_name,
                path=os.path.join(self.pre_p_dir, f_name),
                typemap=TYPEMAP,
            )

        assert "The bus-carrier mapping is empty" in str(e_info.value)

    def test_carrier_under_defined(self):
        """
        If the carriers are not defined within 'elements/bus.csv' file they will be inferred from the components, if the components do not have enough carrier attributes to map all busses an error is raised
        """
        f_name = "carriers_under_defined"
        with pytest.raises(ValueError) as e_info:
            infer_metadata_from_data(
                package_name=f_name,
                path=os.path.join(self.pre_p_dir, f_name),
                typemap=TYPEMAP,
            )

        assert (
            "Bus 'permeate-bus' is missing from the busses carrier dict inferred from the EnergySystem instance"
            in str(e_info.value)
        )

    def test_carrier_wrongly_defined(self):
        """
        If the carriers are not defined within 'elements/bus.csv' file they will be inferred from the components, if two components connected to the same bus have disctinct carriers an error is raised
        """
        f_name = "carriers_wrongly_defined"
        with pytest.raises(ValueError) as e_info:
            infer_metadata_from_data(
                package_name=f_name,
                path=os.path.join(self.pre_p_dir, f_name),
                typemap=TYPEMAP,
            )

        assert "are associated to the same bus" in str(e_info.value)



def pre_processing_moo():
    """This function will run the multi-objective optimization

    The outcome is that the main costs 'capacity_cost' will be replaced by an aggregated
    indicator representing the multi-objective optimization goals.
    This function will replace the pre_processing_costs function if moo is set to True

    Inputs
    Weights: defined by model-user e.g. percentages up to 1
        (in optiMG, the user defining this will be e.g. local prosumers)
    0.5 for costs 0.2 for emissions 0.2 for land requirement 0.1 for water dissipated
    Has to add up to 1, otherwise error

    Normalization: done with global values
    User sets costs are defined in the CSV e.g. CAPEX, OPEX fix, lifetime
        -> this cost is normalised based on global GDP
        -> value is calculated for proportion of cost to global GDP
    Same applies for total emissions
        -> the value is normalised based on total global annual GHG emissions
    Land requirements
        ->  the value is normalised based on the worlds surface area
    Water dissipated
        -> the value is normalised based on e.g. global availability
    Then these values are added together

    User includes specific cost, specific emission factor, specific land requirement, specific water footprint
    This function will take those, normalise based on normalisation data (global)
    One aggregated value will be calculated based on adding these values
    This value will be entered in the csv file under 'capacity_cost'
    The csv file will be updated

    :return:
    """
    return



import os
import shutil

from oemof_tabular_plugins.script import compute_scenario
from oemof_tabular_plugins.datapackage.building import infer_metadata_from_data
from oemof_tabular_plugins.wefe import WEFE_TYPEMAP as TYPEMAP


class TestWaterComponents:
    """
    This class contains tests for the possible ways to use (or not) the mapping of bus to carrier in the results dataframe
    """

    # store the relevant paths
    test_inputs_path = os.path.join("tests", "test_inputs")
    pre_p_dir = os.path.join(
        "tests",
        "test_inputs",
    )
    results_dir = os.path.join("tests", "tests_outputs")
    _package_path = os.path.join("data", "elements")
    package_path = _package_path

    def setup_method(self):
        """
        Sets up the testing environment before each test method is executed.
        """
        if os.path.exists(self.results_dir):
            shutil.rmtree(self.results_dir, ignore_errors=True)
        os.mkdir(self.results_dir)

    def compute_scenario(self, fname, infer_bus_carrier=True):
        """
        Run a single scenario and return a calculator instance
        """
        return compute_scenario(
            os.path.join(self.pre_p_dir, fname),
            self.results_dir,
            wacc=1,
            typemap=TYPEMAP,
            infer_bus_carrier=infer_bus_carrier,
            skip_preprocessing=False,
            skip_infer_datapackage_metadata=True,
        )

    def teardown_method(self):
        """
        Cleans up the testing environment after each test method is executed.
        """
        # removes the 'pre_processing' directory if it exists, ignoring errors
        if os.path.exists(self.results_dir):
            shutil.rmtree(self.results_dir, ignore_errors=True)

    def test_run_water_component_test_through(self):
        """
        The carriers are defined within a column 'carrier' of 'elements/bus.csv' file
        """
        f_name = "water_components"

        calc = self.compute_scenario(f_name)
        assert "carrier" in calc.df_results.index.names



import os
import warnings
import logging
import pandas as pd
from tableschema.exceptions import CastError
from oemof.tabular.datapackage import building
from datapackage import Package

# TODO when oemof.tabular is updated with inferred metadata PR, use
# from oemof.tabular.config import config
import oemof_tabular_plugins.datapackage.config as config

PROFILE_FIELDS = ["profile"]

STR_FIELDS = ["name", "carrier", "tech", "type", "region", "crop_type"]


def map_sequence_profiles_to_resource_name(p, excluded_profiles=("timeindex",)):
    """Look in every resource which is a sequence and map each of its fields to itself

    Within this process the unicity of the field names will be checked, with the exception of the field "timeindex"

    """

    def check_sequences_labels_unicity(labels, new_labels):
        intersect = set(labels).intersection(new_labels)
        if len(intersect) == 1:
            intersect = intersect.pop()
            if not intersect == "timeindex":
                answer = [intersect]
            else:
                answer = []
        else:
            answer = list(intersect)

        if answer:
            warnings.warn(
                f"The labels of the profiles are not unique across all files within 'sequences' folder: '{','.join(intersect)}' used more than once"
            )
        return answer

    sequences = {}
    sequence_labels = []
    duplicated_labels = []
    for r in p.resources:
        if "/sequences/" in r.descriptor["path"]:
            field_labels = [
                f.name for f in r.schema.fields if f.name not in excluded_profiles
            ]
            sequences[r.descriptor["name"]] = field_labels
            duplicated_labels += check_sequences_labels_unicity(
                sequence_labels, field_labels
            )
            sequence_labels += field_labels

    if duplicated_labels:
        # write an error message here
        pass
    # map each profile to its resource name
    sequences_mapping = {
        value: key for (key, values) in sequences.items() for value in values
    }
    return sequences_mapping


def infer_resource_foreign_keys(resource, sequences_profiles_to_resource, busses):
    """Find out the foreign keys within a resource fields

    Look through all field of a resource which are of type 'string' if any of their values are matching a profile header in any of the sequences resources


    Parameters
    ----------
    resource: a :datapackage.Resource: instance
    sequences_profiles_to_resource: the mapping of sequence profile headers to their resource name

    Returns
    -------
    The :datapackage.Resource: instance with updated "foreignKeys" field

    """
    r = resource
    try:
        data = pd.DataFrame.from_records(r.read(keyed=True))
    except CastError:
        raise ValueError(
            f"Error while parsing the resource '{r.descriptor.get('path', r.name)}'. Check that all lines in the file have same number of records. Sometimes it might also a ',' instead of ';' as separator or vice versa."
        )
    # TODO not sure this should be set here
    r.descriptor["schema"]["primaryKey"] = "name"
    if "foreignKeys" not in r.descriptor["schema"]:
        r.descriptor["schema"]["foreignKeys"] = []

    for field in r.schema.fields:
        if field.type == "string" and field.name not in STR_FIELDS:
            for potential_fk in data[field.name].dropna().unique():
                if potential_fk in sequences_profiles_to_resource:
                    # this is actually a wrong format and should be with a "fields" field under the "reference" fields

                    fk = {
                        "fields": field.name,
                        "reference": {
                            "resource": sequences_profiles_to_resource[potential_fk],
                        },
                    }

                    if fk not in r.descriptor["schema"]["foreignKeys"]:
                        r.descriptor["schema"]["foreignKeys"].append(fk)
                elif potential_fk in busses:
                    fk = {
                        "fields": field.name,
                        "reference": {"resource": "bus", "fields": "name"},
                    }
                    if fk not in r.descriptor["schema"]["foreignKeys"]:
                        r.descriptor["schema"]["foreignKeys"].append(fk)
                else:
                    # check for specific fields which are meant to link to profile
                    possible_field_values = [
                        f"'{seq}'" for seq in sequences_profiles_to_resource
                    ]
                    logging.error(
                        f"The value '{potential_fk}' of the field '{field.name}' of the resource '{r.name}' does not match the headers of any sequences. "
                        "If this field is not meant to be a foreign key, you can safely ignore this error :) "
                        f"If this field is meant to be a foreign key to a sequence, then possible values are: {','.join(possible_field_values)}"
                    )

    r.commit()
    return r


def check_profiles(package):
    """Check that values of foreign keys to resources in data/sequences have a match in target sequence headers
    Parameters
    ----------
    package: datapackage instance
    Returns
    -------
    None, raises an error if a value assigned under the foreign key field does not match target sequence headers

    """
    for r in package.resources:
        fkeys = r.descriptor["schema"].get("foreignKeys", [])
        if fkeys:

            data = pd.DataFrame.from_records(r.read(keyed=True))
            for foreign_key in fkeys:
                fk_field = foreign_key["fields"]
                fk_target = foreign_key["reference"]["resource"]
                if fk_target != "bus":
                    for fk_value in data[fk_field].dropna().unique():
                        sequence_descriptor = package.get_resource(fk_target).descriptor
                        sequence_headers = [
                            f"{f['name']}"
                            for f in sequence_descriptor["schema"].get("fields", [])
                        ]
                        if fk_value not in sequence_headers:
                            raise ValueError(
                                f"The value {fk_value} of the field {fk_field} within the resource {r.name} does not match the headers of its resource within '{sequence_descriptor['path']}'\n"
                                f"possible values for the field {fk_field} are: {', '.join(sequence_headers)}"
                            )


def infer_package_foreign_keys(package, typemap=None):
    """Infer the foreign_keys from data/elements and data/sequences and update meta data

    Parameters
    ----------
    package: scenario datapackage
    typemap: facade typemap

    """
    if typemap is None:
        typemap = {}

    p = package
    sequences_profiles_to_resource = map_sequence_profiles_to_resource_name(p)

    bus_data = pd.DataFrame.from_records(p.get_resource("bus").read(keyed=True))

    for r in p.resources:
        # it should be from the folder "elements" and the resource should not be bus.csv
        if (
            "/elements/" in r.descriptor["path"]
            or os.sep + "elements" + os.sep in r.descriptor["path"]
        ) and r.name != "bus":
            r = infer_resource_foreign_keys(
                r, sequences_profiles_to_resource, busses=bus_data.name.to_list()
            )

            if r.name in typemap:
                facade_type = typemap[r.name]
                # test if facade_type has the method 'validate_datapackage'
                if hasattr(facade_type, "validate_datapackage"):
                    # apply the method if it exists
                    r = facade_type.validate_datapackage(r)

            p.remove_resource(r.name)
            p.add_resource(r.descriptor)


def infer_busses_carrier(package, infer_from_component=True):
    """Loop through the nodes of an energy system and infer the carrier of busses from them

    Parameters
    ----------
    package: datapackage.Package instance
    infer_from_component: bool
        if True, the bus carrier mapping will be inferred from the components if not found in 'bus'
        resource of the package

    Returns
    -------
    dict mapping the busses labels to their carrier

    """

    bus_data = pd.DataFrame.from_records(package.get_resource("bus").read(keyed=True))
    bus_carrier = None
    if "carrier" in bus_data.columns:
        bus_carrier = {row[1]["name"]: row[1]["carrier"] for row in bus_data.iterrows()}

    if bus_carrier is None:
        busses_carrier = {}
        if infer_from_component is True:
            for node_type in package.resources:
                if node_type.name != "bus":
                    fields = [f.name for f in node_type.schema.fields]
                    if "carrier" in fields:
                        nodes = node_type.read(keyed=True)
                        for attribute in ("bus", "from_bus", "from_bus_0", "to_bus_1"):
                            for node in nodes:
                                if attribute in node and node["carrier"] != "":

                                    bus_label = node[attribute]
                                    if bus_label in busses_carrier:
                                        if busses_carrier[bus_label] != node["carrier"]:
                                            raise ValueError(
                                                f"Two different carriers ({busses_carrier[bus_label]}, {node['carrier']}) are associated to the same bus '{bus_label}'"
                                            )
                                    else:
                                        busses_carrier[bus_label] = node["carrier"]
        else:
            return
    else:
        logging.info(
            "Bus to carrier mapping found in 'elements/bus.csv' file of datapackage"
        )
        busses_carrier = bus_carrier

    if not busses_carrier:
        raise ValueError(
            "The bus-carrier mapping is empty, this is likely due to missing 'carrier' attributes in the csv files of the elements folder of the datapackage. The simpler way to fix this, is to add a 'carrier' column in the 'elements/bus.csv' file"
        )

    # Check that every bus has a carrier assigned to it
    busses = bus_data.name.to_list()

    for bus_label in busses:
        if bus_label not in busses_carrier:
            print("busses carriers", busses_carrier)
            raise ValueError(
                f"Bus '{bus_label}' is missing from the busses carrier dict inferred from the EnergySystem instance"
            )

    return busses_carrier


def infer_metadata_from_data(
    package_name="default-name",
    path=None,
    metadata_filename="datapackage.json",
    typemap=None,
):
    """

    Returns
    -------

    """

    # Infer the fields from the package data
    path = os.path.abspath(path)
    p0 = Package(base_path=path)
    p0.infer(os.path.join(path, "**" + os.sep + "*.csv"))
    p0.commit()
    p0.save(os.path.join(path, metadata_filename))

    foreign_keys = {}

    def infer_resource_basic_foreign_keys(resource):
        """insert resource foreign_key into a dict formatted for building.infer_metadata

        Compare the fields of a resource to a list of field names known to be foreign keys. If the field name is within the list, it is used to populate the dict 'foreign_keys'
        """

    for r in p0.resources:
        if os.sep + "elements" + os.sep in r.descriptor["path"]:
            infer_resource_basic_foreign_keys(r)
    # this function saves the metadata of the package in json format
    building.infer_metadata(
        package_name=package_name,
        path=path,
        metadata_filename=metadata_filename,
    )

    # reload the package from the saved json file
    p = Package(os.path.join(path, metadata_filename))
    infer_package_foreign_keys(p, typemap=typemap)
    p.descriptor["resources"].sort(key=lambda x: (x["path"], x["name"]))
    p.commit()
    p.save(os.path.join(path, metadata_filename))
    infer_busses_carrier(p)
    check_profiles(p)



"""
Source
------------
SIMPLE crop model: Site information and related soil parameters (Table 1b)
https://doi.org/10.1016/j.eja.2019.01.009

If multiple values for different cultivars are given, the average is used.

Please note that these values are stored as global parameters for every crop,
even though they reflect specific locations which may lead to errors.

Crops
------------
wheat
rice
maize
soybean
drybean
peanut
potato
cassava
tomato
sweetcorn
greenbean
carrot
cotton
banana

Parameters
------------
awc:    average water holding capacity [-]
rcn:    runoff curve number [-]
ddc:    deep drainage coefficient [-]
rzd:    root zone depth [mm]
"""

soil_dict = {
    "banana": {
        "awc": 0.11,
        "rcn": 70,
        "ddc": 0.5,
        "rzd": 1500,
    },
    "cassava": {
        "awc": 0.13,
        "rcn": 85,
        "ddc": 0.4,
        "rzd": 1500,
    },
    "wheat": {
        "awc": 0.135,
        "rcn": 67.5,
        "ddc": 0.4,
        "rzd": 900,
    },
    "tomato": {
        "awc": 0.165,
        "rcn": 75.5,
        "ddc": 0.3,
        "rzd": 900,
    },
    "rice": {
        "awc": 0.12,
        "rcn": 70,
        "ddc": 0.3,
        "rzd": 400,
    },
    "maize": {
        "awc": 0.12,
        "rcn": 70,
        "ddc": 0.3,
        "rzd": 1500,
    },
    "soybean": {
        "awc": 0.13,
        "rcn": 70,
        "ddc": 0.57,
        "rzd": 1630,
    },
    "drybean": {
        "awc": 0.08,
        "rcn": 65,
        "ddc": 0.5,
        "rzd": 700,
    },
    "potato": {
        "awc": 0.11,
        "rcn": 67.25,
        "ddc": 0.625,
        "rzd": 825,
    },
    "sweetcorn": {
        "awc": 0.25,
        "rcn": 60,
        "ddc": 0.7,
        "rzd": 1500,
    },
    "green bean": {
        "awc": 0.08,
        "rcn": 65,
        "ddc": 0.5,
        "rzd": 1000,
    },
    "carrot": {
        "awc": 0.09,
        "rcn": 75,
        "ddc": 0.4,
        "rzd": 500,
    },
    "cotton": {
        "awc": 0.12,
        "rcn": 70,
        "ddc": 0.3,
        "rzd": 600,
    },
}



"""
This subpackage of wefe will contain the global global_specs that can be used across multiple projects
e.g. different PV panel global_specs or different crop global_specs (energy content etc)
ToDo: decide if these global_specs actually belong in oemof-tabular-plugins or if they should be stored somewhere else
"""

from .crop_specs import crop_dict
from .soil_specs import soil_dict



from oemof_tabular_plugins.wefe.constraints.constraint_facades import (
    CONSTRAINT_TYPE_MAP,
)
from oemof.tabular.facades import TYPEMAP
from .facades import WaterPump, WaterFiltration, SimpleCrop, PVPanel, MIMO, MimoCrop

WEFE_TYPEMAP = {
    "water-pump": WaterPump,
    "water-filtration": WaterFiltration,
    "crop": SimpleCrop,
    "mimo-crop": MimoCrop,
    "pv-panel": PVPanel,
    "mimo": MIMO,
}

WEFE_TYPEMAP.update(TYPEMAP)



import argparse
import datetime
import os.path

import pandas as pd
import numpy as np

BASE_PATH = os.path.dirname(os.path.abspath(__file__))

parser = argparse.ArgumentParser(
    prog="oemof-tabular-plugins",
    description="Execute a scenario with oemof-tabular plugins",
    epilog="For more information see ...",
)
parser.add_argument(
    "-i",
    dest="dp_path",
    nargs="+",
    type=str,
    help="Path to the datapackage. Must be provided",
)
parser.add_argument(
    "-o",
    dest="results_path",
    nargs="+",
    type=str,
    help=f"Path to the output folder to save simulation results. If not provided, default output will be provided in {os.path.join(BASE_PATH, 'otp_results')}",
)


parser.add_argument(
    "--start-date",
    dest="date_start",
    type=datetime.date.fromisoformat,
    help="Date of start in YYYY-MM-DD format",
)

parser.add_argument(
    "-p",
    dest="parallel",
    default=False,
    const=True,
    nargs="?",
    type=bool,
    help="Whether or not the simulation uses parallel processing",
)

parser.add_argument(
    "-p",
    dest="parallel",
    default=False,
    const=True,
    nargs="?",
    type=bool,
    help="Whether or not the simulation uses parallel processing",
)


def main():
    args = vars(parser.parse_args())
    fnames = args["fname_path"]
    ofnames = args["ofname_path"]
    num_days = args["num_days"]
    # Define which input files should be considered and run.
    date_start = args["date_start"]
    date_end = args["date_end"]
    ext = args["extension"]
    parallel_processing = args["parallel"]

    years = args["years"]


if __name__ == "__main__":
    main()



import numpy as np
import os
from datapackage import Package
import pandas as pd
import time
import argparse
from datetime import datetime, timedelta
from oemof import solph
from oemof.tools.economics import annuity

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

try:
    from oemof_visio import ESGraphRenderer

    ES_GRAPH = True
except ModuleNotFoundError:
    ES_GRAPH = False
z_version = 1


import dash
from dash import dcc
from dash import html
from dash import dash_table
from dash.dependencies import Input, Output, State
import plotly.graph_objs as go

RESULTS_COLUMN_NAMES = [
    "annuity",
    "annual_costs",
    "total_flow",
    "capacity",
    "cash_flow",  # AA: could be named fuel_expenditure_cost
    "total_opex_costs",
    "first_investment",
]

service__item_style = {
    "border-style": "solid",
    "border-width": "3px",
    "padding": "1rem",
    "margin": "1rem",
    "border-radius": "5px",
}

table__item_style = {
    "kpis": {
        "width": "500px",
        "border-style": "solid",
        "border-width": "3px",
        "padding": "1rem",
        "margin": "1rem",
        "border-radius": "5px",
    },
    "capacities": {
        "width": "800px",
        "border-style": "solid",
        "border-width": "3px",
        "padding": "1rem",
        "margin": "1rem",
        "border-radius": "5px",
    },
}

container_style = {
    "display": "flex",
    "flex-direction": "row",
    "flex-wrap": "wrap",
    "justify-content": "flex-start",
}


##########################################################################
# Initialize the energy system and calculate necessary parameters
##########################################################################


def encode_image_file(img_path):
    """Encode image files to load them in the dash layout under img html tag

    Parameters
    ----------
    img_path: str
        path to the image file

    Returns
    -------
    encoded_img: bytes
        encoded bytes of the image file

    """

    try:
        with open(img_path, "rb") as ifs:
            encoded_img = base64.b64encode(ifs.read())
    except FileNotFoundError:
        encoded_img = base64.b64encode(bytes())
    return encoded_img


def sankey(energy_system, ts=None):
    """Return a dict to a plotly sankey diagram"""
    busses = []

    labels = []
    sources = []
    targets = []
    values = []

    results = energy_system.results

    # draw a node for each of the network's component. The shape depends on the component's type
    for nd in energy_system.nodes:
        if isinstance(nd, solph.Bus):

            # keep the bus reference for drawing edges later
            bus = nd
            busses.append(bus)

            bus_label = bus.label

            labels.append(nd.label)

            flows = solph.views.node(results, bus_label)["sequences"]

            # draw an arrow from the component to the bus
            for component in bus.inputs:
                if component.label not in labels:
                    labels.append(component.label)

                sources.append(labels.index(component.label))
                targets.append(labels.index(bus_label))

                try:
                    val = flows[
                        (
                            (component.label, bus_label),
                            (component.label, bus_label, "flow"),
                        )
                    ].sum()
                except Exception as e:
                    val = flows[((component.label, bus_label), "flow")].sum()

                if ts is not None:
                    try:
                        val = flows[
                            (
                                (component.label, bus_label),
                                (component.label, bus_label, "flow"),
                            )
                        ][ts]
                    except:
                        val = flows[((component.label, bus_label), "flow")][ts]
                # if val == 0:
                #     val = 1
                values.append(val)

            for component in bus.outputs:
                # draw an arrow from the bus to the component
                if component.label not in labels:
                    labels.append(component.label)

                sources.append(labels.index(bus_label))
                targets.append(labels.index(component.label))

                try:
                    val = flows[
                        (
                            (bus_label, component.label),
                            (bus_label, component.label, "flow"),
                        )
                    ].sum()
                except Exception as e:
                    val = flows[((bus_label, component.label), "flow")].sum()
                if ts is not None:
                    try:
                        val = flows[
                            (
                                (bus_label, component.label),
                                (bus_label, component.label, "flow"),
                            )
                        ][ts]
                    except:
                        val = flows[
                            (
                                (bus_label, component.label),
                                "flow",
                            )
                        ][ts]
                values.append(val)

    fig = go.Figure(
        data=[
            go.Sankey(
                node=dict(
                    pad=15,
                    thickness=20,
                    line=dict(color="black", width=0.5),
                    label=labels,
                    hovertemplate="Node has total value %{value}<extra></extra>",
                    color="blue",
                ),
                link=dict(
                    source=sources,  # indices correspond to labels, eg A1, A2, A2, B1, ...
                    target=targets,
                    value=values,
                    hovertemplate="Link from node %{source.label}<br />"
                    + "to node%{target.label}<br />has value %{value}"
                    + "<br />and data <extra></extra>",
                ),
            )
        ]
    )

    fig.update_layout(title_text="Basic Sankey Diagram", font_size=10)
    return fig.to_dict()


def prepare_app(energy_system, dp_path, tables, services, units=None):

    # TODO to display energy system
    energy_system_graph = f"energy_system.png"
    # if ES_GRAPH is True:
    #     es = ESGraphRenderer(
    #     energy_system, legend=True, filepath=energy_system_graph, img_format="png"
    #     )
    #     es.render()
    #     energy_system_graph = encode_image_file(f"energy_system.png")

    results = energy_system.results

    bus_figures = []

    p0 = Package(dp_path)
    bus_data = pd.DataFrame.from_records(p0.get_resource("bus").read(keyed=True))
    busses = bus_data.name.tolist()

    date_time_index = energy_system.timeindex

    for bus in busses:
        if bus != "battery":
            fig = go.Figure(layout=dict(title=f"{bus} bus node"))
            for t, g in solph.views.node(results, node=bus)["sequences"].items():
                idx_asset = abs(t[0].index(bus) - 1)

                fig.add_trace(
                    go.Scatter(
                        x=g.index, y=g.values * pow(-1, idx_asset), name=t[0][idx_asset]
                    )
                )
        # else:
        #     capacity_battery = asset_results.capacity.battery
        #     if capacity_battery != 0:
        #         soc_battery = solph.views.node(results, node=bus)["sequences"][
        #                           (("battery", "None"), "storage_content")] / capacity_battery
        #     else:
        #         soc_battery = solph.views.node(results, node=bus)["sequences"][
        #             (("battery", "None"), "storage_content")]
        #
        #     fig = go.Figure(layout=dict(title=f"{bus} node"))
        #
        #     fig.add_trace(
        #         go.Scatter(
        #             x=soc_battery.index, y=soc_battery.values, name="soc battery"
        #         )
        #     )

        bus_figures.append(fig)

    tables_figure = []

    for table in tables:
        df = tables[table]

        if "unit" not in df.columns:
            df.reset_index(inplace=True)

            def set_value(row_number, assigned_value):
                return assigned_value.get(row_number, None)

            df["unit"] = df[df.columns[0]].apply(set_value, args=(units,))

        tables_figure.append(
            html.Div(
                style=table__item_style[table],
                children=[
                    html.H4(table),
                    dash_table.DataTable(
                        data=df.round(2).to_dict("records"),
                        columns=[{"name": i, "id": i} for i in df.columns],
                        style_cell_conditional=[
                            {"if": {"column_id": "kpi"}, "textAlign": "center"},
                            {
                                "if": {"column_id": "Component name"},
                                "textAlign": "center",
                            },
                        ],
                    ),
                ],
            )
        )

    services_figure = []
    for service in services:
        df = services[service]
        carrier = df.carrier.unique()[0]
        unit = units.get(carrier, "UNIT NOT FOUND")

        is_crop = False

        if "crop" in df.facade_type.values:
            is_crop = True

        df = df.rename(
            columns={"asset": "Component name", "aggregated_flow": unit},
        )

        # Populate the subtables of the service
        production = df.loc[df.direction == "out"].copy()

        if is_crop is False:
            usage = df.loc[(df.direction == "in") & (df.facade_type != "excess")].copy()
            excess = df.loc[df.facade_type == "excess"].copy()
            tables = [production, usage, excess]
        else:
            usage = df.loc[df.direction == "in"].copy()
            tables = [production, usage]

        for table in tables:
            if table.empty is False:
                table.drop(
                    columns=["carrier", "direction", "facade_type"], inplace=True
                )
                table.loc[:, "Percentage"] = 100 * table[unit] / table[unit].sum()
                # Add a line with "total" if there is more than one component
                if len(table) > 1:
                    summary_line = table.iloc[:, 1:].sum()
                    summary_line["Component name"] = "Total"
                    table.loc[-1] = summary_line

        table_headers = ["Production", "Usage"]
        if is_crop is False:
            if excess.empty is False:
                if excess[unit].sum() > 0:
                    table_headers.append("Excess")

        services_figure.append(
            html.Div(
                id=f"{service}-service-div",
                className="service--item",
                style=service__item_style,
                children=[
                    html.H4(service.replace("-", " ").capitalize()),
                    html.Div(
                        children=[
                            html.Div(
                                [
                                    html.H5(table_hdr),
                                    dash_table.DataTable(
                                        data=table.round(2).to_dict("records"),
                                        columns=[
                                            {"name": i, "id": i} for i in table.columns
                                        ],
                                        style_cell_conditional=[
                                            {
                                                "if": {"column_id": i},
                                                "width": f"{100/len(table.columns)}%",
                                            }
                                            for i in table.columns
                                        ]
                                        + [
                                            {
                                                "if": {"column_id": "Component name"},
                                                "textAlign": "center",
                                            }
                                        ],
                                    ),
                                ]
                            )
                            for table_hdr, table in zip(table_headers, tables)
                        ],
                    ),
                ],
            )
        )

    # loading external resources
    external_stylesheets = ["https://codepen.io/chriddyp/pen/bWLwgP.css"]
    options = dict(
        # external_stylesheets=external_stylesheets
    )

    demo_app = dash.Dash(__name__, **options)

    demo_app.layout = html.Div(
        children=[
            html.H2("Scalar results"),
            html.H3("KPIS"),
            html.Div(
                className="table--container",
                style=container_style,
                children=tables_figure,
            ),
            html.H3("Services"),
            # dcc.Dropdown(
            #     options=[s for s in services],
            #     value=[s for s in services],
            #     id="service_select",
            #     multi=True
            # ),
            dcc.Checklist(
                id="service_select",
                options=[s for s in services],
                value=[],  # s for s in services],
                inline=True,
            ),
            html.Div(
                services_figure,
                className="service--container",
                style=container_style,
            ),
            html.H2("Dynamic results"),
            html.P(
                children=[
                    "You can adjust the slider to get the energy flow at a single timestep, "
                    "or look for a specific timestep in the dropdown menu below ",
                    html.Span(
                        "Note if you change the slider "
                        "it will show the value in the dropdown menu, but it you change the dropdown menu directly "
                        "it will not update the slider)"
                    ),
                ]
            ),
            dcc.Slider(
                id="ts_slice_slider",
                value=1,
                min=0,
                max=len(date_time_index),
                # marks={k: v for k, v in enumerate(date_time_index)},
            ),
            dcc.Dropdown(
                id="ts_slice_select",
                options={k: v for k, v in enumerate(date_time_index)},
                value=None,
            ),
            dcc.Graph(id="sankey", figure=sankey(energy_system)),
        ]
        + [
            dcc.Graph(
                id=f"{bus}-id",
                figure=fig,
            )
            for bus, fig in zip(busses, bus_figures)
        ]
        + [dcc.Graph(id="sankey_aggregate", figure=sankey(energy_system))]
        # + [
        #     html.H4(["Energy system"]),
        #     html.Img(
        #         src="data:image/png;base64,{}".format(energy_system_graph.decode()),
        #         alt="Energy System Graph, if you do not see this image it is because pygraphviz is not installed. "
        #             "If you are a windows user it might be complicated to install pygraphviz.",
        #         style={"maxWidth": "100%"},
        #     ),
        # ]
    )

    @demo_app.callback(
        # The value of these components of the layout will be changed by this callback
        [
            Output(component_id="sankey", component_property="figure"),
        ]
        + [
            Output(component_id=f"{bus}-id", component_property="figure")
            for bus in busses
        ],
        # Triggers the callback when the value of one of these components of the layout is changed
        Input(component_id="ts_slice_select", component_property="value"),
    )
    def update_figures(ts):
        if ts is None:
            ts = "0"
        ts = int(ts)
        # see if case changes, otherwise do not rerun this
        date_time_index = energy_system.timeindex

        bus_figures = []
        for bus in busses:
            if bus != "battery":
                fig = go.Figure(layout=dict(title=f"{bus} bus node"))
                max_y = 0
                for t, g in solph.views.node(results, node=bus)["sequences"].items():
                    idx_asset = abs(t[0].index(bus) - 1)
                    asset_name = t[0][idx_asset]
                    if t[0][idx_asset] == "battery":
                        if idx_asset == 0:
                            asset_name += " discharge"
                        else:
                            asset_name += " charge"
                    opts = {}
                    negative_sign = pow(-1, idx_asset)
                    opts["stackgroup"] = (
                        "negative_sign" if negative_sign < 0 else "positive_sign"
                    )

                    fig.add_trace(
                        go.Scatter(
                            x=g.index,
                            y=g.values * negative_sign,
                            name=asset_name,
                            **opts,
                        )
                    )
                    if g.max() > max_y:
                        max_y = g.max()
            else:
                capacity_battery = asset_results.capacity.battery
                if capacity_battery != 0:
                    soc_battery = (
                        solph.views.node(results, node=bus)["sequences"][
                            (("battery", "None"), "storage_content")
                        ]
                        / capacity_battery
                    )
                else:
                    soc_battery = solph.views.node(results, node=bus)["sequences"][
                        (("battery", "None"), "storage_content")
                    ]

                fig = go.Figure(layout=dict(title=f"{bus} node", yaxis_range=[0, 1]))

                fig.add_trace(
                    go.Scatter(
                        x=soc_battery.index, y=soc_battery.values, name="soc battery"
                    )
                )
                fig.add_trace(
                    go.Scatter(
                        x=soc_battery.index,
                        y=np.ones(len(soc_battery.index)) * settings.storage_soc_min,
                        name="min soc battery",
                    )
                )
                fig.add_trace(
                    go.Scatter(
                        x=soc_battery.index,
                        y=np.ones(len(soc_battery.index)) * settings.storage_soc_max,
                        name="max soc battery",
                    )
                )
            fig.add_trace(
                go.Scatter(
                    x=[date_time_index[ts], date_time_index[ts]],
                    y=[0, max_y],
                    name="none",
                    line_color="black",
                )
            )
            bus_figures.append(fig)

        return [
            sankey(energy_system, date_time_index[ts]),
        ] + bus_figures

    @demo_app.callback(
        # The value of these components of the layout will be changed by this callback
        Output(component_id="ts_slice_select", component_property="value"),
        # Triggers the callback when the value of one of these components of the layout is changed
        Input(component_id="ts_slice_slider", component_property="value"),
    )
    def change_ts_value(val):
        return val

    @demo_app.callback(
        # The value of these components of the layout will be changed by this callback
        [
            Output(component_id=f"{s}-service-div", component_property="style")
            for s in services
        ],
        # Triggers the callback when the value of one of these components of the layout is changed
        Input(component_id="service_select", component_property="value"),
    )
    def change_visibility_value(val):
        answer = [
            (
                service__item_style | {"display": "block"}
                if s in val
                else service__item_style | {"display": "none"}
            )
            for s in services
        ]
        return answer

    return demo_app

    # import ipdb;ipdb.set_trace()



import logging
import os
from datapackage import Package
import warnings

import pandas as pd
from oemof.tabular.postprocessing import calculations as clc, naming
from oemof.tabular.postprocessing.core import Calculator


from oemof_tabular_plugins.datapackage.post_processing import (
    construct_dataframe_from_results,
    process_raw_results,
    process_raw_inputs,
    apply_calculations,
    apply_kpi_calculations,
    RAW_INPUTS,
    RAW_OUTPUTS,
    PROCESSED_RAW_OUTPUTS,
    CALCULATED_OUTPUTS,
    CALCULATED_KPIS,
)

from .gui import prepare_app

# ------ New post-processing to create tables ------
# This dictionary contains groups of columns that should be extracted from the df_results to generate a clearer overview


# TODO add a column for planned capacity (not optimizable but including costs) in capacities if it gets properly
#  implemented (planned capacity can be set by setting capacity_minimum == capacity_potential and dispatchable = True
RESULT_TABLE_COLUMNS = {
    "costs": ["upfront_investment_costs", "annuity_total", "variable_costs_total"],
    "capacities": [
        "capacity",
        "storage_capacity",
        "capacity_minimum",
        "capacity_potential",
        "storage_capacity_potential",
        "expandable",
        "investments",
        "capacity_total",
    ],
}

CAPACITIES_UNIT = {
    "electricity": {"default": "[kW]", "storage": "[kWh]"},
    "wind": {"default": "[m³/s]", "storage": "[m³/s]"},
    "irradiation": {"default": "[m²]", "storage": "[m²]"},
    "crop": {"default": "[m²]", "storage": "[m²]"},
    "water": {"default": "[m³/s]", "storage": "[m³]"},
}


def extract_table_from_results(df_results, columns):
    """Extracts a set of columns from the df_results DataFrame. The lists of columns to generate these tables can be
    defined in RESULT_TABLE_COLUMNS
    :param df_results: multiindex results dataframe with additional columns (OTPCalculator.df_results)
    :param columns: list of columns to generate the sub-table (should be defined in RESULT_TABLE_COLUMNS)
    :return: dataframe containing the columns specified in columns, if present in df_results
    """
    missing_columns = []
    for col in columns:
        try:
            df_results[col]
        except KeyError:
            # If the key is not in the df_results dataframe, log it as a warning and pop the column from the list
            logging.warning(
                f"The column {col} was not found in the results DataFrame, will be skipped in the subtable"
            )
            missing_columns.append(col)

    columns = [col for col in columns if col not in missing_columns]
    results_table = df_results[columns].copy()
    # TODO some of these names may be confusing because they are just the columns, maybe there should be a
    #  verbose parameter in CALCULATED OUTPUTS that we can then also use here
    results_table.columns = [
        col.title().replace("_", " ") for col in results_table.columns
    ]
    return results_table


def save_table_to_csv(table, results_path, filename):
    """Saves a DataFrame to a .csv file"""
    filepath = os.path.join(results_path, filename)
    table.to_csv(filepath)


# TODO figure out the best table/display for storage results


# --------------------------------------------------
class OTPCalculator(Calculator):
    def __init__(
        self, input_parameters, energy_system, dp_path, infer_bus_carrier=True
    ):

        self.df_results = construct_dataframe_from_results(
            energy_system, dp_path=dp_path, infer_bus_carrier=infer_bus_carrier
        )
        self.n_timesteps = len(energy_system.timeindex)

        self.df_results = process_raw_results(self.df_results)
        self.df_results = process_raw_inputs(self.df_results, dp_path)
        self.kpis = None

        try:
            super().__init__(input_parameters, energy_system.results)
        except Exception as e:
            logging.error(
                f"Initialisation of Calculator did not succeed: {str(e)}. This should not impact oemof-tabular-plugins results."
            )

    def apply_calculations(self, calculations):
        apply_calculations(self.df_results, calculations=calculations)

    def apply_kpi_calculations(self, calculations):
        self.kpis = apply_kpi_calculations(self.df_results, calculations=calculations)

    def __scalars(self, scalar_category):
        """Ignore the flow data columns (by construction those are the first columns after the multi-index)"""
        scalars = self.df_results.iloc[:, self.n_timesteps :]
        answer = scalars
        if scalar_category == "raw_inputs":
            existing_cols = []
            for c in scalars.columns:
                if c in RAW_INPUTS:
                    existing_cols.append(c)
            answer = scalars[existing_cols]
        elif scalar_category == "outputs":
            answer = scalars[scalars.columns.difference(RAW_INPUTS)]
        return answer

    @property
    def raw_outputs(self):
        self.df_results.iloc[:, : self.n_timesteps]
        cols = self.df_results.iloc[:, : self.n_timesteps].columns.tolist()
        cols = cols + RAW_OUTPUTS + PROCESSED_RAW_OUTPUTS
        return self.df_results[cols]

    @property
    def raw_inputs(self):
        return self.__scalars("raw_inputs")

    @property
    def calculated_outputs(self):
        return self.__scalars("outputs")


def post_processing(
    params,
    es,
    results_path,
    dp_path,
    dash_app=False,
    parameters_units=None,
    infer_bus_carrier=True,
    calculations=None,
    kpi_calculations=None,
):
    # ToDo: adapt this function after multi-index dataframe is implemented to make it more concise / cleaner
    # ToDo: params can be accessed in results so will not need to be a separate argument
    """
    The main post-processing function extracts various scalar and timeseries data and stores it in CSV files.
    :param params: energy system parameters
    :param es: oemof energy_system with results in it, ie es.results = processing.results(m) has been performed
    :param results_path: results directory path
    :param dp_path: path to the datapackage.json file
    """

    if parameters_units is None:
        #  Units of Capacities and Kpis in Results
        parameters_units = {
            "battery_storage": "[kWh]",
            "inverter": "[kW]",
            "pv-panel": "[kW]",
            "water-storage": "[m³]",
            "mimo": "[m³/h]",
            "annuity_total": "[$]",
            "total_upfront_investments": "[$]",
            "land_requirement_total": "[m²]",
            "total_water_footprint": "[m³]",
        }

    if calculations is None:
        calculations = CALCULATED_OUTPUTS
    if kpi_calculations is None:
        kpi_calculations = CALCULATED_KPIS
    # initiate calculator for post-processing
    calculator = OTPCalculator(params, es, dp_path, infer_bus_carrier=infer_bus_carrier)
    calculator.apply_calculations(calculations)
    calculator.apply_kpi_calculations(kpi_calculations)

    tables_to_save = {}

    results_by_flow = calculator.df_results

    result_tables = {}
    services_table = {}

    if results_by_flow is not None:
        results_by_flow.to_csv(results_path + "/all_results_by_flow.csv", index=True)
        # get sub-tables from results dataframe
        capacities_table = extract_table_from_results(
            results_by_flow, RESULT_TABLE_COLUMNS["capacities"]
        )

        # ignore the dispatchable sources optimized capacities
        capacities_table = capacities_table[
            capacities_table.index.get_level_values("facade_type") != "dispatchable"
        ]

        # Assign units to the capacities:
        def set_unit(carrier, facade_type):
            unit_choice = CAPACITIES_UNIT.get(carrier, {"default": "", "storage": ""})

            if facade_type in unit_choice:
                unit = unit_choice[facade_type]
            else:
                unit = unit_choice["default"]
            return unit

        capacities_table.reset_index(inplace=True)
        if "Capacity Potential" not in capacities_table.columns:
            capacities_table["Capacity Potential"] = None

        if "carrier" not in capacities_table.columns:
            capacities_table["carrier"] = None

        capacities_table["unit"] = capacities_table.apply(
            lambda x: set_unit(x.carrier, x.facade_type), axis=1
        )

        capacities_table.rename(
            columns={
                "asset": "Component name",
                "Investments": "Optimized Capacity",
                "Capacity Potential": "Maximum Capacity",
            },
            inplace=True,
        )

        # eliminate double occurences of same asset
        capacities_table = capacities_table.loc[
            (capacities_table["Capacity Total"] > 0)
            & (capacities_table.direction == "out"),
            [
                "Component name",
                "Capacity",
                "Optimized Capacity",
                "Capacity Total",
                "Maximum Capacity",
                "unit",
            ],
        ]
        result_tables.update({"capacities": capacities_table})

        cost_table = extract_table_from_results(
            calculator.df_results, RESULT_TABLE_COLUMNS["costs"]
        )

        # result_tables.update({"costs": cost_table})

        # IDEA define calculations for WEFE components in post_processing and make a merge within __init__ of WEFE
        # TODO list components (facades) automatically (low prio)
        df = results_by_flow.reset_index()
        if "carrier" not in df.columns:
            df["carrier"] = None
        service_busses = df.loc[df.facade_type == "load"].bus.tolist()
        service_busses += df.loc[
            (df.direction == "out") & (df.facade_type == "crop")
        ].bus.tolist()

        for bus in service_busses:
            df_bus = df.loc[df.bus == bus]

            services_table[bus.replace("-bus", "")] = df_bus[
                ["asset", "direction", "aggregated_flow", "carrier", "facade_type"]
            ]

        # save tables to csv files
        tables_to_save.update(
            {"costs.csv": cost_table, "capacities.csv": capacities_table}
        )
    kpis = calculator.kpis
    if kpis is not None:
        kpis.to_csv(results_path + "/kpis.csv", index=True)

        if "mimo" in results_by_flow.index.get_level_values("asset"):
            kpis.loc["total_water_produced"] = results_by_flow.loc[
                "permeate-bus", "in"
            ]["aggregated_flow"].sum()
            kpis.loc["total_brine_produced"] = results_by_flow.loc["brine-bus", "in"][
                "aggregated_flow"
            ].sum()
            kpis.loc["total_electricity_produced"] = results_by_flow.loc[
                "ac-elec-bus", "in"
            ]["aggregated_flow"].sum()

        result_tables.update({"kpis": kpis})

    for filename, table in tables_to_save.items():
        save_table_to_csv(table, results_path, filename)

    if dash_app is True:

        demo_app = prepare_app(
            es,
            dp_path=dp_path,
            tables=result_tables,
            services=services_table,
            units=parameters_units,
        )
        demo_app.run_server(debug=False, port=8060)

    # ----- OLD POST-PROCESSING - TO BE DELETED ONCE CERTAIN -----
    if hasattr(calculator, "scalar_params"):
        try:
            # calculate scalars using functions from clc module
            aggregated_flows = clc.AggregatedFlows(calculator).result
            storage_losses = clc.StorageLosses(calculator).result
            transmission_losses = clc.TransmissionLosses(calculator).result
            invested_capacity = clc.InvestedCapacity(calculator).result
            invested_storage_capacity = clc.InvestedStorageCapacity(calculator).result
            invested_capacity_costs = clc.InvestedCapacityCosts(calculator).result
            invested_storage_capacity_costs = clc.InvestedStorageCapacityCosts(
                calculator
            ).result
            summed_carrier_costs = clc.SummedCarrierCosts(calculator).result
            summed_marginal_costs = clc.SummedMarginalCosts(calculator).result
            total_system_costs = clc.TotalSystemCosts(calculator).result

            # combine all results into a single dataframe
            all_scalars = [
                aggregated_flows,
                storage_losses,
                transmission_losses,
                invested_capacity,
                invested_storage_capacity,
                invested_capacity_costs,
                invested_storage_capacity_costs,
                summed_carrier_costs,
                summed_marginal_costs,
            ]
            # map variable names and add component information
            all_scalars = pd.concat(all_scalars, axis=0)
            all_scalars = naming.map_var_names(
                all_scalars,
                calculator.scalar_params,
                calculator.busses,
                calculator.links,
            )
            all_scalars = naming.add_component_info(all_scalars, calculator.scalar_params)
            print("Total System Cost", total_system_costs)
            total_system_costs.index.names = ("name", "var_name")
            all_scalars = pd.concat([all_scalars, total_system_costs], axis=0)
            all_scalars = all_scalars.sort_values(by=["carrier", "tech", "var_name"])
            # save all scalar results to a csv file
            filepath_name_all_scalars = os.path.join(results_path, "all_scalars.csv")
            all_scalars.to_csv(filepath_name_all_scalars)

            # saves all hourly timeseries as a dataframe (see test_postprocessing.py in oemof-tabular/tests
            # for example if wanting to filter particular nodes)
            all_sequences = clc.AggregatedFlows(calculator, resample_mode="H")
            # save all timeseries (sequence) results to a csv file
            filepath_name_all_sequences = os.path.join(results_path, "all_sequences.csv")
            all_sequences.sequences.to_csv(filepath_name_all_sequences)
        except Exception as e:
            logging.error(e)

    return calculator



from .constraints.constraint_facades import CONSTRAINT_TYPE_MAP
from .post_processing.post_processing import post_processing
from .pre_processing.pre_processing import pre_processing
from .pre_processing.pre_processing import calculate_annuity
from .pre_processing.pre_processing import logger
from .post_processing.gui import prepare_app



import os
import pandas as pd
import logging
from oemof.tools import logger, economics
import json
from .pre_processing_moo import pre_processing_moo

logger.define_logging()


def calculate_annuity(capex, opex_fix, lifetime, wacc):
    """
    Calculates the total annuity for each component, including CAPEX and fixed OPEX.
    :param capex: CAPEX (currency/MW*) *or the unit you choose to use throughout the model e.g. kW/GW
    :param opex_fix: fixed OPEX (currency/MW*/year)
    :param lifetime: lifetime of the component (years)
    :param wacc: weighted average cost of capital (WACC) applied throughout the model (%)
    :return: total annuity (currency/MW*/year)
    """
    annuity_capex = economics.annuity(capex, lifetime, wacc)
    annuity_opex_fix = opex_fix
    annuity_total = round(annuity_capex + annuity_opex_fix, 2)
    return annuity_total


def pre_processing_costs(wacc, element, element_path, element_df):
    """
    Applies pre-processing costs to the input CSV files, where the annuity ('capacity_cost') is either
    used directly if stated, or if left empty then calculated using the calculate_annuity function,
    or if all parameters are stated a choice is given.
    :param wacc: weighted average cost of capital (WACC) applied throughout the model (%)
    :param element: csv filename
    :param element_path: path of the csv file
    :param element_df: dataframe containing data from the csv file
    """
    # for every element other than storage, the annuity cost parameter is 'capacity_cost'
    # for storage, the annuity cost parameter is 'storage_capacity_cost'
    if element != "storage.csv":
        annuity_cost = "capacity_cost"
    else:
        annuity_cost = "storage_capacity_cost"
    # check if any of the required columns are missing
    cost_columns = {"capex", "opex_fix", "lifetime"}
    missing_columns = cost_columns - set(element_df.columns)

    # ---------------- POSSIBLE SCENARIO GROUPS FOR PARAMETER ENTRIES ----------------
    # scenario group "annuity no cost params": annuity parameter is included and all of capex, opex fix
    # and lifetime parameters are not included in the csv file
    if annuity_cost in element_df.columns and cost_columns == missing_columns:
        scenario_group = "annuity no cost params"
    # scenario group "annuity partial cost params": annuity parameter is included and some but not all of
    # capex, opex fix and lifetime parameters are included in the csv file
    elif (
        annuity_cost in element_df.columns
        and missing_columns
        and cost_columns != missing_columns
    ):
        scenario_group = "annuity partial cost params"
    # scenario group "annuity all cost params": annuity parameter is included and all of capex, opex fix
    # and lifetime parameters are included in the csv file
    elif annuity_cost in element_df.columns and not missing_columns:
        scenario_group = "annuity all cost params"
    # scenario group "no annuity partial/all cost params": annuity parameter is not included and at least one
    # of capex, opex fix and lifetime parameters are included in the csv file
    elif annuity_cost not in element_df.columns and cost_columns != missing_columns:
        scenario_group = "no annuity partial/all cost params"
    # scenario group "no annuity no cost params": the annuity parameter is not included and neither are
    # capex, opex fix and lifetime parameters in the csv file
    elif annuity_cost not in element_df.columns and cost_columns == missing_columns:
        scenario_group = "no annuity no cost params"

    # ---------------- POSSIBLE SCENARIOS FOR EACH SCENARIO GROUP ----------------
    # loop through each entry in the csv file
    for index, row in element_df.iterrows():
        # define the row name
        row_name = row["name"]
        if scenario_group == "annuity no cost params":
            # scenario "annuity empty no cost params": the annuity parameter is left empty and the other cost
            # parameters have not been included
            if pd.isna(row[annuity_cost]):
                scenario = "annuity empty no cost params"
            # scenario "annuity defined no cost params": the annuity parameter is defined and the other cost
            # parameters have not been included
            else:
                scenario = "annuity defined no cost params"
        elif scenario_group == "annuity partial cost params":
            # scenario "annuity empty partial cost params": the annuity parameter is left empty and only
            # some other financial parameters are included
            if pd.isna(row[annuity_cost]):
                scenario = "annuity empty partial cost params"
            # scenario "annuity defined partial cost params": the annuity parameter is defined and only
            # some other financial parameters are included
            else:
                scenario = "annuity defined partial cost params"
        elif scenario_group == "annuity all cost params":
            # store the parameters
            capex = row["capex"]
            opex_fix = row["opex_fix"]
            lifetime = row["lifetime"]
            # scenario "annuity empty all cost params": the annuity parameter is left empty and all of the
            # other financial parameters are defined
            if (
                pd.isna(row[annuity_cost])
                and pd.notna(capex)
                and pd.notna(opex_fix)
                and pd.notna(lifetime)
            ):
                scenario = "annuity empty all cost params defined"
            # scenario "annuity all cost params some empty": the annuity parameter is either defined or empty,
            # but at least one of 'capex', 'opex_fix' and 'lifetime' is left empty
            elif pd.isna(capex) or pd.isna(opex_fix) or pd.isna(lifetime):
                scenario = "annuity all cost params some empty"
            # scenario "annuity defined all cost params defined": both the annuity parameter is defined and
            # all the other financial parameters are defined
            else:
                scenario = "annuity defined all cost params defined"
        elif scenario_group == "no annuity partial/all cost params":
            # store the parameters
            capex = row["capex"]
            opex_fix = row["opex_fix"]
            lifetime = row["lifetime"]
            # scenario "no annuity partial/all cost params empty": at least one of 'capex', 'opex_fix' and
            # 'lifetime' is left empty
            if pd.isna(capex) or pd.isna(opex_fix) or pd.isna(lifetime):
                scenario = "no annuity partial/all cost params empty"
            # scenario "no annuity all cost params defined": all financial parameters are defined
            else:
                scenario = "no annuity all cost params defined"
        elif scenario_group == "no annuity no cost params":
            # scenario "no annuity no cost params": neither the annuity or financial parameters are defined
            scenario = "no annuity no cost params"

        # ---------------- ACTIONS TAKEN FOR EACH SCENARIO ----------------
        if scenario == "annuity empty no cost params":
            # raise value error
            raise ValueError(
                f"'{annuity_cost}' (the annuity) has been left empty for '{row_name}' "
                f"in '{element}', and 'capex', 'opex_fix' and 'lifetime' have not "
                f" been included. \nEither the annuity ('{annuity_cost}') must be "
                f"directly stated or all of the other financial parameters must be stated "
                f"to calculate the annuity."
            )
        elif scenario == "annuity defined no cost params":
            # log info message
            logger.info(
                f"The annuity cost is directly used for '{row_name}' in '{element}'."
            )
        elif scenario == "annuity empty partial cost params":
            # raise value error
            raise ValueError(
                f"'{annuity_cost}' (the annuity) has been left empty for '{row_name}' "
                f"in '{element}', and not all of 'capex', 'opex_fix' and 'lifetime' have"
                f" been included. \nEither the annuity ('{annuity_cost}') must be "
                f"directly stated or all of the other financial parameters must be stated "
                f"to calculate the annuity."
            )
        elif scenario == "annuity defined partial cost params":
            # log warning message
            logging.warning(
                f"'{annuity_cost}' (the annuity) has been defined and some but not all "
                f"of 'capex', 'opex_fix' and 'lifetime' have been defined for {row_name} "
                f"in {element}. The annuity will be directly used but be aware that some "
                f"cost results will not be calculated."
            )
        elif scenario == "annuity empty all cost params defined":
            # calculate the annuity using the calculate_annuity function
            capacity_cost = calculate_annuity(capex, opex_fix, lifetime, wacc)
            # update the dataframe
            element_df.at[index, annuity_cost] = float(capacity_cost)
            # log info message
            logger.info(
                f"the annuity ('{annuity_cost}') has been calculated and updated for"
                f" '{row_name}' in '{element}'."
            )
        elif scenario == "annuity all cost params some empty":
            # log warning message
            logging.warning(
                f"One or more of 'capex', 'opex_fix' and 'lifetime' have been left "
                f"empty for {row_name} in {element}. The annuity will be directly used "
                f"but be aware that some cost results will not be calculated."
            )
        elif scenario == "annuity defined all cost params defined":
            capacity_cost = calculate_annuity(capex, opex_fix, lifetime, wacc)
            # update the dataframe
            element_df.at[index, annuity_cost] = float(capacity_cost)
            # if all parameters are defined, the user is asked if they want to calculate the annuity
            # from the capex, opex_fix and lifetime or use the annuity directly
            logger.info(
                f"All parameters ('capex', 'opex_fix', 'lifetime') and '{annuity_cost}' are "
                f"provided for '{row_name}' in '{element}'. \nThe defined annuity has been replaced with "
                f"the calculated value from capex, opex_fix and lifetime."
            )
        elif scenario == "no annuity partial/all cost params empty":
            # raise value error
            raise ValueError(
                f"One or more of 'capex', 'opex_fix' and 'lifetime' have been left "
                f"empty for {row_name} in {element}. Please enter values or remove the"
                f" parameters and include "
                f"the 'capacity_cost'."
            )
        elif scenario == "no annuity all cost params defined":
            # calculate the annuity using the calculate_annuity function
            capacity_cost = calculate_annuity(capex, opex_fix, lifetime, wacc)
            # update the dataframe
            element_df["capacity_cost"] = float(capacity_cost)
            # log info message
            logger.info(
                f"the annuity ('{annuity_cost}') has been calculated and updated for"
                f" '{row_name}' in '{element}'."
            )
        elif scenario == "no annuity no cost params":
            logger.info(
                f"'{element}' does not contain '{annuity_cost}' parameter. Skipping..."
            )

    # save the updated dataframe to the csv file
    element_df.to_csv(element_path, sep=";", index=False)
    return


def pre_processing_custom_attributes(element_path, element_df, custom_attributes):
    # ToDo: confirm if this function is needed, whether the attributes need to be added to
    #  'output_parameters' or if this is not necessary
    """Updates the 'output_parameters' field in the CSV file for the specified element if custom
    attributes are defined.

    :param element_path: path of the csv file
    :param element_df: dataframe containing data from the csv file
    :param custom_attributes: list of custom attributes included in the model (defined in compute.py)
    """
    # iterate over each entry in the dataframe (from csv file)
    for index, row in element_df.iterrows():
        # create empty custom attributes dict
        custom_attributes_dict = {}
        # set boolean to false
        has_custom_attributes = False
        # check if custom_attributes is not none before iterating
        if custom_attributes is not None:
            # check if any of the custom attributes list are in the dataframe columns
            for attribute in custom_attributes:
                if attribute in element_df.columns:
                    value = row[attribute]
                    # add the attribute to the custom attributes dict
                    custom_attributes_dict[attribute] = value
                    # set boolean to true
                    has_custom_attributes = True
                # if custom attributes are found for this row, add them to 'output_parameters'
                if has_custom_attributes:
                    output_parameters_str = json.dumps(
                        {"custom_attributes": custom_attributes_dict}
                    )
                    element_df.at[index, "output_parameters"] = output_parameters_str
                else:
                    # no custom attributes found, do not update 'output_parameters'
                    continue
    # write the updated dataframe back to the csv file
    element_df.to_csv(element_path, sep=";", index=False)
    return


def pre_processing(scenario_dir, wacc, custom_attributes=None, moo=False):
    """Performs pre-processing of input scenario data before running the model.

    :param scenario_dir: scenario directory path
    :param wacc: weighted average cost of capital (WACC) applied throughout the model (%)
    :param custom_attributes: list of custom attributes included in the model (defined in compute.py), default is None
    :param moo: whether the multi-objective optimization is activated, default is False
    """
    if moo is False:
        logger.info(f"Optimization activated for only costs")
    elif moo is True:
        logger.info(f"Multi-objective optimization activated")
    logger.info("Pre-processing activated")
    # locate the elements directory
    elements_dir = os.path.join(scenario_dir, "data", "elements")
    # raise error if the elements directory is not found in the scenario directory
    if not os.path.exists(elements_dir):
        raise FileNotFoundError(f"No 'elements' directory found in {scenario_dir}.")
    # loop through each csv file in the elements directory
    for element in os.listdir(elements_dir):
        # only consider csv files
        if element.endswith(".csv"):
            # set the path of the considered csv file
            element_path = os.path.join(elements_dir, element)
            # read the csv file and save it as a pandas dataframe
            element_df = pd.read_csv(element_path, sep=";")
            if moo is False:
                # performs pre-processing of additional cost data (capex, opex_fix, lifetime)
                pre_processing_costs(wacc, element, element_path, element_df)
                # performs pre-processing for custom attributes (e.g. emission factor, renewable factor, land
                # requirement)
                pre_processing_custom_attributes(
                    element_path, element_df, custom_attributes
                )
            elif moo is True:
                pre_processing_moo()
    logger.info("Pre-processing completed")
    return



import logging

try:
    from oemof_industry.mimo_converter import MIMO
except ModuleNotFoundError:
    logging.error(
        "oemof industry is not installed, install it with 'pip install git+https://github.com/sedos-project/oemof.industry.git@saltwater'"
    )
    from oemof.tabular.facades import Conversion

    MIMO = Conversion


def validate_mimo_datapackage(cls, resource):
    """
    This datapackage validation method is necessary for the MIMO converter because the
    'primary' field automatically gets updated to the foreign keys in the datapackage.json
    by the infer_metadata_from_data, but it should not be interpreted as a foreign key.
    This method removes it.

    :param cls: class instance
    :param resource: the datapackage resource
    """
    # check if the 'foreignKeys' field exists in the schema
    if (
        "schema" in resource.descriptor
        and "foreignKeys" in resource.descriptor["schema"]
    ):
        # loop through each foreign key
        for foreign_key in resource.descriptor["schema"]["foreignKeys"]:
            if "primary" in foreign_key["fields"]:
                # remove the foreign_key regarding 'primary' from the resource
                resource.descriptor["schema"]["foreignKeys"].remove(foreign_key)
                break
    return resource


def processing_mimo_raw_inputs(cls, resource, results_df):
    # function to apply on df from above (drop the thee columns (conversion_factor_ac-elec-bus,
    # conversion_factor_permeate-bus, conversion_factor_brine-bus) and turn them into one
    # conversion_factor column), then add the primary column

    return results_df


MIMO.validate_datapackage = classmethod(validate_mimo_datapackage)
MIMO.processing_raw_inputs = classmethod(processing_mimo_raw_inputs)
